import datetime

import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.formatters.datetime import DatetimeFormatter


@pytest.fixture
def datetime_test_df():
    row_cnt = 1000
    header = [
        "int_id",
        "str_id",
        "not_int_id",
        "not_str_id",
        "simple_datetime",
        "simple_datetime_2",
        "date_with_time",
    ]

    int_id = list(range(row_cnt))
    str_id = list("id_" + str(i) for i in range(row_cnt))

    not_int_id = list(range(int(row_cnt / 2))) + list(range(int(row_cnt / 2)))
    not_str_id = list("id_" + str(i) for i in range(int(row_cnt / 2))) + list(
        "id_" + str(i) for i in range(int(row_cnt / 2))
    )

    simple_datetime = pd.date_range(start="2023-12-27", periods=row_cnt)

    # such "%d %b %Y" cannot be directly marked as a datetime column by pandas
    # but can be identified through the to_datetime method, which is implemented by DatetimeInspector
    simple_datetime_2 = [datetime.datetime.strftime(x, "%d %b %Y") for x in simple_datetime]
    simple_datetime_str = [datetime.datetime.strftime(x, "%Y-%m-%d") for x in simple_datetime]

    h = np.random.randint(0, 24, size=row_cnt)
    m = np.random.randint(0, 59, size=row_cnt)
    s = np.random.randint(0, 59, size=row_cnt)
    date_with_time = [
        simple_datetime[i] + pd.Timedelta(hours=h[i], minutes=m[i], seconds=s[i])
        for i in range(row_cnt)
    ]

    datetime_test_df = [
        [
            int_id[i],
            str_id[i],
            not_int_id[i],
            not_str_id[i],
            simple_datetime_str[i],
            simple_datetime_2[i],
            date_with_time[i],
        ]
        for i in range(row_cnt)
    ]

    yield pd.DataFrame(datetime_test_df, columns=header)  # 1000 rows, 7 columns


def is_an_integer_list(lst):
    """
    Check if all elements in the list are integers or floats that are also integers.

    Parameters:
    lst (list): The list to be checked.

    Returns:
    bool: True if all elements are integers or floats that are also integers, False otherwise.
    """
    return all(isinstance(i, int) or (isinstance(i, float) and i.is_integer()) for i in lst)


def is_a_string_list(lst):
    """
    Check if all items in a list are strings.

    Parameters:
    lst (list): The list to check.

    Returns:
    bool: True if all items in the list are strings, False otherwise.
    """
    return all(isinstance(item, str) for item in lst)


def test_datetime_formatter_test_df_dead_column(datetime_test_df: pd.DataFrame):
    """
    Test the DatetimeFormatter class with a DataFrame that has datetime columns.

    Parameters:
    datetime_test_df (pd.DataFrame): The DataFrame to test.

    Returns:
    None

    Raises:
    AssertionError: If any of the assertions fail.
    """
    # about the data
    assert datetime_test_df.shape == (1000, 7)

    # get the metadata
    metadata_df = Metadata.from_dataframe(datetime_test_df)
    assert metadata_df.datetime_columns == {
        "simple_datetime_2",
        "date_with_time",
        "simple_datetime",
    }
    metadata_df.datetime_format = {}

    # fit the transformer
    transformer = DatetimeFormatter()
    transformer.fit(metadata=metadata_df)

    # no element in datetime_columns
    assert transformer.datetime_columns == []
    assert set(transformer.dead_columns) == {
        "simple_datetime_2",
        "date_with_time",
        "simple_datetime",
    }  # all dead


def test_datetime_formatter_test_df(datetime_test_df: pd.DataFrame):
    """
    Test function for the DatetimeFormatter class.

    This function tests the functionality of the DatetimeFormatter class by creating a test DataFrame,
    setting the datetime format for the columns, fitting the transformer, converting the DataFrame,
    reversing the conversion, and checking if the reversed DataFrame is equal to the original one.

    Args:
        datetime_test_df (pd.DataFrame): The test DataFrame to be used for testing.

    Returns:
        None

    Raises:
        AssertionError: If any of the assertions fail.
    """
    # about the data
    assert datetime_test_df.shape == (1000, 7)

    # get the metadata
    metadata_df = Metadata.from_dataframe(datetime_test_df)
    assert metadata_df.datetime_columns == {
        "simple_datetime_2",
        "date_with_time",
        "simple_datetime",
    }
    # set the right format
    # TODO : it seems that metadata has problems in inspecting defalut format such as "%d %b %Y"
    # here we assign the real value, and will fix this issue in another PR
    datetime_format = {}
    datetime_format["simple_datetime"] = "%Y-%m-%d"
    datetime_format["simple_datetime_2"] = "%d %b %Y"
    datetime_format["date_with_time"] = "%Y-%m-%d %H:%M:%S"
    metadata_df.datetime_format = datetime_format

    # fit the transformer
    transformer = DatetimeFormatter()
    assert not transformer.fitted
    transformer.fit(metadata=metadata_df)
    assert transformer.fitted

    # no element in dead_columns
    assert transformer.dead_columns == []  # no columns dead
    assert set(transformer.datetime_columns) == {
        "simple_datetime_2",
        "date_with_time",
        "simple_datetime",
    }

    # convert the dataframe, check if datetime columns are int type
    converted_df = transformer.convert(datetime_test_df)
    assert is_an_integer_list(converted_df["date_with_time"].to_list())
    assert is_an_integer_list(converted_df["simple_datetime_2"].to_list())
    assert is_an_integer_list(converted_df["simple_datetime"].to_list())

    reverse_converte_df = transformer.reverse_convert(converted_df)
    # assert string type in reverse converted dataframe
    assert is_a_string_list(reverse_converte_df["simple_datetime"].to_list())
    assert is_a_string_list(reverse_converte_df["date_with_time"].to_list())
    assert is_a_string_list(reverse_converte_df["simple_datetime_2"].to_list())

    # check if the dataframe is equal to the original one
    # use the eq method and .all().all() to check the equality of two 2d dataframes
    assert reverse_converte_df.eq(datetime_test_df).all().all()



import datetime
import itertools
import random
import string
from pathlib import Path

import click

_HERE = Path(__file__).parent


def random_string(length):
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def random_float():
    return random.random() * 1000


def random_int():
    return random.randint(0, 1000)


def random_timestamp():
    current_timestamp = datetime.datetime.now().timestamp()
    return current_timestamp + random_int()


def random_datetime():
    return datetime.datetime.fromtimestamp(random_timestamp())


@click.option(
    "--output_file",
    default=(_HERE / "dataset/benchmark.csv").as_posix(),
)
@click.option("--num_rows", default=1_000_000)
@click.option("--int_cols", default=15)
@click.option("--float_cols", default=15)
@click.option("--string_cols", default=10)
@click.option("--string_discrete_nums", default=50)
@click.option("--timestamp_cols", default=10)
@click.option("--datetime_cols", default=0)
@click.command()
def generate_dateset(
    output_file,
    num_rows,
    int_cols,
    float_cols,
    string_cols,
    string_discrete_nums,
    timestamp_cols,
    datetime_cols,
):
    headers = itertools.chain.from_iterable(
        [
            (f"int_col{i}" for i in range(int_cols)),
            (f"float_col{i}" for i in range(float_cols)),
            (f"string_col{i}" for i in range(string_cols)),
            (f"timestamp_col{i}" for i in range(timestamp_cols)),
            (f"datetime_col{i}" for i in range(datetime_cols)),
        ]
    )
    output_file = Path(output_file).expanduser().resolve()
    output_file.parent.mkdir(parents=True, exist_ok=True)

    random_str_list = [random_string(25) for i in range(string_discrete_nums)]

    def _generate_one_line():
        return ",".join(
            map(
                str,
                itertools.chain(
                    (random_int() for _ in range(int_cols)),
                    (random_float() for _ in range(float_cols)),
                    (random.choice(random_str_list) for _ in range(string_cols)),
                    (random_timestamp() for _ in range(timestamp_cols)),
                    (random_datetime() for _ in range(datetime_cols)),
                ),
            )
        )

    with output_file.open("w") as f:
        f.write(",".join(headers) + "\n")

        chunk_size = 1000
        for i in range(0, num_rows, chunk_size):
            f.write("\n".join(_generate_one_line() for _ in range(chunk_size)))
            f.write("\n")


if __name__ == "__main__":
    generate_dateset()



from __future__ import annotations

from typing import Generator

import pandas as pd


class DataConnector:
    """
    DataConnector warps data source into ``pd.DataFrame``.

    For different data source, implement a specific subclass.
    """

    identity = None
    """
    Identity of data source, e.g. table name, hash of content
    """

    def _read(self, offset: int = 0, limit: int | None = None) -> pd.DataFrame | None | None:
        """
        Subclass must implement this for reading data.

        See ``read`` for more details.
        """
        raise NotImplementedError

    def _columns(self) -> list[str]:
        """
        Subclass should implement this for reading columns if there is an efficient way for peaking columns.

        See ``column`` for more details.
        """
        raise NotImplementedError

    def _iter(self, offset: int = 0, chunksize: int = 0) -> Generator[pd.DataFrame, None, None]:
        """
        Subclass should implement this for reading data in chunk.

        See ``iter`` for more details.
        """
        raise NotImplementedError

    def iter(self, offset: int = 0, chunksize: int = 0) -> Generator[pd.DataFrame, None, None]:
        """
        Interface for reading data in chunk.

        Args:
            offset (int, optional): Offset for reading. Defaults to 0.
            chunksize (int, optional): Chunksize for reading. Defaults to 0.

        Returns:
            typing.Generator[pd.DataFrame, None, None]: Generator/Iterator for readed dataframe
        """
        return self._iter(offset, chunksize)

    def read(self, offset: int = 0, limit: int | None = None) -> pd.DataFrame | None:
        """
        Interface for reading data.

        Args:
            offset (int, optional): Offset for reading. Defaults to 0.
            limit (int, optional): Limit for reading. Defaults to None.
                None is for reading all data and 0 is for reading no data(only header).

        Returns:
            pd.DataFrame: Readed dataframe
        """
        return self._read(offset, limit)

    def columns(self) -> list[str]:
        """
        Interface for peaking columns.
        """
        try:
            return self._columns()
        except NotImplementedError:
            return self.read(0, 1).columns.tolist()

    def keys(self) -> list[str]:
        """
        Same as ``columns``.
        """
        return self.columns()

    def finalize(self):
        """
        Finalize the data connector.
        """
        pass



from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from typing import Any, Dict

import numpy as np
import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.formatters.base import Formatter
from sdgx.utils import logger


class DatetimeFormatter(Formatter):
    """
    A class for formatting datetime columns in a pandas DataFrame.

    DatetimeFormatter is designed to handle the conversion of datetime columns to timestamp format and vice versa.
    It uses metadata to identify datetime columns and their corresponding datetime formats.

    Attributes:
        datetime_columns (list): List of column names that are of datetime type.
        datetime_formats (dict): Dictionary with column names as keys and datetime formats as values.
        dead_columns (list): List of column names that are no longer needed or to be removed.
        fitted (bool): Indicates whether the formatter has been fitted.

    Methods:
        fit(metadata: Metadata | None = None, **kwargs: dict[str, Any]): Fits the formatter by recording the datetime columns and their formats.
        convert(raw_data: pd.DataFrame) -> pd.DataFrame: Converts datetime columns in raw_data to timestamp format.
        reverse_convert(processed_data: pd.DataFrame) -> pd.DataFrame: Converts timestamp columns in processed_data back to datetime format.
    """

    datetime_columns: list
    """
    List to store the columns that are of datetime type.
    """

    datetime_formats: Dict
    """
    Dictionary to store the datetime formats for each column, with default value as an empty string.
    """

    dead_columns: list
    """
    List to store columns that are no longer needed or to be removed.
    """

    def __init__(self):
        self.fitted = False
        self.datetime_columns = []
        self.datetime_formats = defaultdict(str)
        self.dead_columns = []

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for datetime formatter, the datetime column and datetime format need to be recorded.

        If there is a column without format, the default format will be used for output (this may cause some problems).

        Formatter need to use metadata to record which columns belong to datetime type, and convert timestamp back to datetime type during post-processing.
        """

        # get from metadata
        self.datetime_formats = metadata.get("datetime_format")
        datetime_columns = []
        dead_columns = []
        # Check datetime_formats and columns
        # exclude columns without format as there is huge risk of handling errors
        meta_datetime_columns = metadata.get("datetime_columns")
        for each_col in meta_datetime_columns:
            if each_col in self.datetime_formats.keys():
                datetime_columns.append(each_col)
            else:
                dead_columns.append(each_col)
                logger.warning(
                    f"Column {each_col} has no datetime_format, DatetimeFormatter will REMOVE this column！"
                )

        # Remove successful formatted datetime columns from metadata.discrete_columns
        if not (set(datetime_columns) - set(metadata.discrete_columns)):
            metadata.change_column_type(datetime_columns, "discrete", "datetime")
        # Remove dead_columns from metadata
        metadata.remove_column(dead_columns)

        self.datetime_columns = datetime_columns
        self.dead_columns = dead_columns

        logger.info("DatetimeFormatter Fitted.")
        self.fitted = True
        return

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method to convert datetime samples into timestamp.

        Args:
            - raw_data (pd.DataFrame): Unprocessed table data
        """
        if len(self.datetime_columns) == 0:
            logger.info(
                "Converting data using DatetimeFormatter... Finished (No datetime columns)."
            )
            return raw_data

        # remove the column without format
        for each_col in self.dead_columns:
            raw_data = self.remove_columns(raw_data, [each_col])
            logger.warning(f"Column {each_col} was removed because lack of format info.")

        logger.info("Converting data using DatetimeFormatter...")

        res_data = self.convert_datetime_columns(
            self.datetime_columns, self.datetime_formats, raw_data
        )

        logger.info("Converting data using DatetimeFormatter... Finished.")

        return res_data

    @staticmethod
    def convert_datetime_columns(datetime_column_list, datetime_formats, processed_data):
        """
        Convert datetime columns in processed_data from string to timestamp (int)

        Args:
            - datetime_column_list (list): List of columns that are date time type
            - processed_data (pd.DataFrame): Processed table data

        Returns:
            - result_data (pd.DataFrame): Processed table data with datetime columns converted to timestamp
        """

        def datetime_formatter(each_value, datetime_format):
            """
            convert each single column datetime string to timestamp int value.
            """
            try:
                datetime_obj = datetime.strptime(str(each_value), datetime_format)
                each_stamp = datetime.timestamp(datetime_obj)
            except Exception as e:
                logger.warning(
                    f"An error occured when convert str to timestamp {e}, we set as mean."
                )
                logger.warning(f"Input parameters: ({str(each_value)}, {datetime_format})")
                logger.warning(f"Input type: ({type(each_value)}, {type(datetime_format)})")
                each_stamp = np.nan
            return each_stamp

        # Make a copy of processed_data to avoid modifying the original data
        result_data: pd.DataFrame = processed_data.copy()

        # Convert each datetime column in datetime_column_list to timestamp
        for column in datetime_column_list:
            # Convert datetime to timestamp (int)
            result_data[column] = result_data[column].apply(
                datetime_formatter, datetime_format=datetime_formats[column]
            )
            result_data[column].fillna(result_data[column].mean(), inplace=True)
        return result_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        reverse_convert method for datetime formatter.

        Does not require any action.
        """

        if len(self.datetime_columns) == 0:
            logger.info("Data reverse-converted by DatetimeFormatter (No datetime columns).")
            return processed_data

        logger.info("Data reverse-converting by DatetimeFormatter...")

        logger.info(f"parameters : {self.datetime_columns}, {self.datetime_formats}")

        result_data = self.convert_timestamp_to_datetime(
            self.datetime_columns, self.datetime_formats, processed_data
        )

        logger.info("Data reverse-converted by DatetimeFormatter... Finished.")

        return result_data

    @staticmethod
    def convert_timestamp_to_datetime(timestamp_column_list, format_dict, processed_data):
        """
        Convert timestamp columns to datetime format in a DataFrame.

        Parameters:
            - timestamp_column_list (list): List of column names in the DataFrame which are of timestamp type.
            - datetime_column_dict (dict): Dictionary with column names as keys and datetime format as values.
            - processed_data (pd.DataFrame): DataFrame containing the processed data.

        Returns:
            - result_data (pd.DataFrame): DataFrame with timestamp columns converted to datetime format.

        TODO:
            if the value <0, the result will be `No Datetime`, try to fix it.
        """

        def column_timestamp_formatter(each_stamp: int, timestamp_format: str) -> str:
            try:
                each_str = datetime.fromtimestamp(each_stamp).strftime(timestamp_format)
            except Exception as e:
                logger.debug(f"An error occured when convert timestamp to str {e}.")
                each_str = "No Datetime"
            return each_str

        # Copy the processed data to result_data
        result_data = processed_data.copy()

        # Iterate over each column in the timestamp_column_list
        for column in timestamp_column_list:
            # Check if the column is in the DataFrame
            if column in result_data.columns:
                # Convert the timestamp to datetime format using the format provided in datetime_column_dict
                result_data[column] = result_data[column].apply(
                    column_timestamp_formatter, timestamp_format=format_dict[column]
                )
            else:
                logger.error(f"Column {column} not in processed data's column list!")

        return result_data


@hookimpl
def register(manager):
    manager.register("DatetimeFormatter", DatetimeFormatter)



"""Dataset Generators for boolean transformers."""

from abc import ABC

import numpy as np

from sdgx.models.components.sdv_rdt.performance.datasets.base import (
    BaseDatasetGenerator,
)

MAX_PERCENT_NULL = 50  # cap the percentage of null values at 50%
MIN_PERCENT = 20  # the minimum percentage of true or false is 20%


class BooleanGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate boolean data."""

    SDTYPE = "boolean"


class RandomBooleanGenerator(BooleanGenerator):
    """Generator that creates dataset of random booleans."""

    @staticmethod
    def generate(num_rows):
        """Generate an array of random booleans.

        Args:
            num_rows (int):
                Number of rows of booleans to generate.

        Returns:
            numpy.ndarray of size ``num_rows`` containing random booleans.
        """
        return np.random.choice(a=[True, False], size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 500.0,
            },
        }


class RandomBooleanNaNsGenerator(BooleanGenerator):
    """Generator that creates an array of random booleans with nulls."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        percent_null = np.random.randint(MIN_PERCENT, MAX_PERCENT_NULL)
        percent_true = (100 - percent_null) / 2
        percent_false = 100 - percent_true - percent_null

        return np.random.choice(
            a=[True, False, None],
            size=num_rows,
            p=[percent_true / 100, percent_false / 100, percent_null / 100],
        )

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 1000.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 1000.0,
            },
        }


class RandomSkewedBooleanGenerator(BooleanGenerator):
    """Generator that creates dataset of random booleans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        percent_true = np.random.randint(MIN_PERCENT, 100 - MIN_PERCENT)

        return np.random.choice(
            a=[True, False],
            size=num_rows,
            p=[percent_true / 100, (100 - percent_true) / 100],
        )

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 500.0,
            },
        }


class RandomSkewedBooleanNaNsGenerator(BooleanGenerator):
    """Generator that creates an array of random booleans with nulls."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        percent_null = np.random.randint(MIN_PERCENT, MAX_PERCENT_NULL)
        percent_true = np.random.randint(MIN_PERCENT, 100 - percent_null - MIN_PERCENT)
        percent_false = 100 - percent_null - percent_true

        return np.random.choice(
            a=[True, False, None],
            size=num_rows,
            p=[percent_true / 100, percent_false / 100, percent_null / 100],
        )

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 1000.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 1000.0,
            },
        }


class ConstantBooleanGenerator(BooleanGenerator):
    """Generator that creates a constant array with either True or False."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = np.random.choice([True, False])
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 500.0,
            },
        }


class ConstantBooleanNaNsGenerator(BooleanGenerator):
    """Generator that creates a constant array with either True or False with some nulls."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = np.random.choice([True, False])
        percent_null = np.random.randint(MIN_PERCENT, MAX_PERCENT_NULL)

        return np.random.choice(
            a=[constant, None],
            size=num_rows,
            p=[(100 - percent_null) / 100, percent_null / 100],
        )

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-5, "memory": 400.0},
            "transform": {"time": 1e-5, "memory": 1000.0},
            "reverse_transform": {
                "time": 5e-5,
                "memory": 1000.0,
            },
        }



import datetime

import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.inspectors.datetime import DatetimeInspector


@pytest.fixture
def inspector():
    yield DatetimeInspector()


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


@pytest.fixture
def datetime_test_df():
    row_cnt = 1000
    header = [
        "int_id",
        "str_id",
        "not_int_id",
        "not_str_id",
        "simple_datetime",
        "simple_datetime_2",
        "date_with_time",
    ]

    int_id = list(range(row_cnt))
    str_id = list("id_" + str(i) for i in range(row_cnt))

    not_int_id = list(range(int(row_cnt / 2))) + list(range(int(row_cnt / 2)))
    not_str_id = list("id_" + str(i) for i in range(int(row_cnt / 2))) + list(
        "id_" + str(i) for i in range(int(row_cnt / 2))
    )

    simple_datetime = pd.date_range(start="2023-12-27", periods=row_cnt)

    # such "%d %b %Y" cannot be directly marked as a datetime column by pandas
    # but can be identified through the to_datetime method, which is implemented by DatetimeInspector
    simple_datetime_2 = [datetime.datetime.strftime(x, "%d %b %Y") for x in simple_datetime]

    h = np.random.randint(0, 24, size=row_cnt)
    m = np.random.randint(0, 59, size=row_cnt)
    s = np.random.randint(0, 59, size=row_cnt)
    date_with_time = [
        simple_datetime[i] + pd.Timedelta(hours=h[i], minutes=m[i], seconds=s[i])
        for i in range(row_cnt)
    ]

    X = [
        [
            int_id[i],
            str_id[i],
            not_int_id[i],
            not_str_id[i],
            simple_datetime[i],
            simple_datetime_2[i],
            date_with_time[i],
        ]
        for i in range(row_cnt)
    ]

    yield pd.DataFrame(X, columns=header)


def test_inspector_demo_data(inspector: DatetimeInspector, raw_data):
    inspector.fit(raw_data)
    assert inspector.ready
    # should be empty set
    assert not inspector.datetime_columns
    assert sorted(inspector.inspect()["datetime_columns"]) == sorted([])
    assert inspector.inspect_level == 20


def test_inspector_generated_data(inspector: DatetimeInspector, datetime_test_df: pd.DataFrame):
    # use generated id data
    inspector.fit(datetime_test_df)
    assert inspector.datetime_columns
    assert sorted(inspector.inspect()["datetime_columns"]) == sorted(
        ["simple_datetime", "simple_datetime_2", "date_with_time"]
    )
    assert inspector.inspect_level == 20


def test_custom_format_detection(datetime_test_df: pd.DataFrame):
    # Instantiate the DatetimeInspector with the custom formats
    inspector = DatetimeInspector(user_formats=["%Y-%m-%d %H:%M:%S"])
    inspector.fit(datetime_test_df)

    # Get the detected datetime formats
    result = inspector.inspect()

    # Assert that the detected formats are correct
    assert result["datetime_formats"]["simple_datetime"] == "%Y-%m-%d %H:%M:%S"
    assert result["datetime_formats"]["simple_datetime_2"] == "%d %b %Y"
    assert result["datetime_formats"]["date_with_time"] == "%Y-%m-%d %H:%M:%S"
    assert inspector.inspect_level == 20


def test_detect_datetime_format_partial_and_no_match(inspector):
    partial_match_series = pd.Series(["2023-03-17", "invalid-date"])
    no_match_series = pd.Series(["not-a-date"])

    assert inspector.detect_datetime_format(partial_match_series) == None
    assert inspector.detect_datetime_format(no_match_series) == None
    assert inspector.inspect_level == 20


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""Personal Identifiable Information Transformers module."""

from sdgx.models.components.sdv_rdt.transformers.pii.anonymizer import (
    AnonymizedFaker,
    PseudoAnonymizedFaker,
)

__all__ = [
    "AnonymizedFaker",
    "PseudoAnonymizedFaker",
]



"""Dataset Generators for categorical transformers."""

from abc import ABC

import numpy as np

from sdgx.models.components.sdv_rdt.performance.datasets.base import (
    BaseDatasetGenerator,
)
from sdgx.models.components.sdv_rdt.performance.datasets.datetime import (
    RandomGapDatetimeGenerator,
)
from sdgx.models.components.sdv_rdt.performance.datasets.utils import add_nans


class CategoricalGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate catgorical data."""

    SDTYPE = "categorical"


class RandomIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array of random integers."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = [1, 2, 3, 4, 5]
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 5e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 1000.0,
            },
        }


class RandomIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random integers with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 5e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 1000.0,
            },
        }


class RandomStringGenerator(CategoricalGenerator):
    """Generator that creates an array of random strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = ["Alice", "Bob", "Charlie", "Dave", "Eve"]
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 500.0},
            "transform": {"time": 1e-05, "memory": 500.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 1000.0,
            },
        }


class RandomStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomStringGenerator.generate(num_rows).astype("O"))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 1e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 1000.0,
            },
        }


class RandomMixedGenerator(CategoricalGenerator):
    """Generator that creates an array of random mixed types.

    Mixed types include: int, float, bool, string, datetime.
    """

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        cat_size = 5
        categories = np.hstack(
            [
                cat.astype("O")
                for cat in [
                    RandomGapDatetimeGenerator.generate(cat_size),
                    np.random.randint(0, 100, cat_size),
                    np.random.uniform(0, 100, cat_size),
                    np.arange(cat_size).astype(str),
                    np.array([True, False]),
                ]
            ]
        )

        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 1e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 2000.0,
            },
        }


class RandomMixedNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random mixed types with nans.

    Mixed types include: int, float, bool, string, datetime.
    """

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        array = RandomMixedGenerator.generate(num_rows)

        length = len(array)
        num_nulls = np.random.randint(1, length)
        nulls_idx = np.random.choice(range(length), num_nulls)
        nulls = np.random.choice([np.nan, float("nan"), None], num_nulls)
        array[nulls_idx] = nulls

        return array

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 1e-05, "memory": 2000.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 2000.0,
            },
        }


class SingleIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array with a single integer."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = np.random.randint(0, 100)
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 3e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 400.0,
            },
        }


class SingleIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array with a single integer with some nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(SingleIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 3e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 500.0,
            },
        }


class SingleStringGenerator(CategoricalGenerator):
    """Generator that creates an array of a single string."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = "A"
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 4e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 400.0,
            },
        }


class SingleStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of a single string with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(SingleStringGenerator.generate(num_rows).astype("O"))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 400.0},
            "transform": {"time": 3e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 500.0,
            },
        }


class UniqueIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array of unique integers."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.arange(num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 2000.0},
            "transform": {"time": 0.0003, "memory": 500000.0},
            "reverse_transform": {
                "time": 0.0004,
                "memory": 1000000.0,
            },
        }


class UniqueIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of unique integers with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(UniqueIntegerGenerator.generate(num_rows))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 1000.0},
            "transform": {"time": 0.0002, "memory": 1000000.0},
            "reverse_transform": {
                "time": 0.0002,
                "memory": 1000000.0,
            },
        }


class UniqueStringGenerator(CategoricalGenerator):
    """Generator that creates an array of unique strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.arange(num_rows).astype(str)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 2000.0},
            "transform": {"time": 0.0002, "memory": 500000.0},
            "reverse_transform": {
                "time": 0.0004,
                "memory": 1000000.0,
            },
        }


class UniqueStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of unique strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(UniqueStringGenerator.generate(num_rows).astype("O"))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 2e-05, "memory": 1000.0},
            "transform": {"time": 0.0005, "memory": 1000000.0},
            "reverse_transform": {
                "time": 0.0002,
                "memory": 1000000.0,
            },
        }



"""BaseSynthesizer module."""

import contextlib
import warnings
from pathlib import Path
from typing import Union

import cloudpickle
import numpy as np
import torch


@contextlib.contextmanager
def set_random_states(random_state, set_model_random_state):
    """Context manager for managing the random state.

    Args:
        random_state (int or tuple):
            The random seed or a tuple of (numpy.random.RandomState, torch.Generator).
        set_model_random_state (function):
            Function to set the random state on the model.
    """
    original_np_state = np.random.get_state()
    original_torch_state = torch.get_rng_state()

    random_np_state, random_torch_state = random_state

    np.random.set_state(random_np_state.get_state())
    torch.set_rng_state(random_torch_state.get_state())

    try:
        yield
    finally:
        current_np_state = np.random.RandomState()
        current_np_state.set_state(np.random.get_state())
        current_torch_state = torch.Generator()
        current_torch_state.set_state(torch.get_rng_state())
        set_model_random_state((current_np_state, current_torch_state))

        np.random.set_state(original_np_state)
        torch.set_rng_state(original_torch_state)


def random_state(function):
    """Set the random state before calling the function.

    Args:
        function (Callable):
            The function to wrap around.
    """

    def wrapper(self, *args, **kwargs):
        if self.random_states is None:
            return function(self, *args, **kwargs)

        else:
            with set_random_states(self.random_states, self.set_random_state):
                return function(self, *args, **kwargs)

    return wrapper


class BaseSynthesizer:
    """Base class for all default synthesizers of ``CTGAN``.

    This should contain the save/load methods.
    """

    random_states = None

    def __getstate__(self):
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        state = self.__dict__.copy()
        self.set_device(device_backup)

        random_states = self.random_states
        if (
            isinstance(random_states, tuple)
            and isinstance(random_states[0], np.random.RandomState)
            and isinstance(random_states[1], torch.Generator)
        ):
            state["_numpy_random_state"] = random_states[0].get_state()
            state["_torch_random_state"] = random_states[1].get_state()
            del state["random_states"]

        return state

    def __setstate__(self, state):
        np_state = state.pop("_numpy_random_state", None)
        torch_state = state.pop("_torch_random_state", None)
        if np_state is not None and torch_state is not None:
            current_torch_state = torch.Generator()
            current_torch_state.set_state(torch_state)
            current_numpy_state = np.random.RandomState()
            current_numpy_state.set_state(np_state)
            state["random_states"] = (current_numpy_state, current_torch_state)
        self.__dict__ = state

    def set_device(self, device):
        """Set the `device` to be used ('GPU' or 'CPU')."""
        self._device = device
        if self._generator is not None:
            self._generator.to(self._device)

    def save(self, path):
        """Save the model in the passed `path`."""
        device_backup = self._device
        self.set_device(torch.device("cpu"))
        with open(path, "wb") as output:
            cloudpickle.dump(self, output)
        self.set_device(device_backup)

    @classmethod
    def load(
        cls, path: Union[str, Path], device: str = "cuda" if torch.cuda.is_available() else "cpu"
    ):
        """Load the model stored in the passed arg `path`."""
        with open(path, "rb") as f:
            model = cloudpickle.load(f)
        model.set_device(device)
        return model

    def set_random_state(self, random_state):
        """Set the random state.

        Args:
            random_state (int, tuple, or None):
                Either a tuple containing the (numpy.random.RandomState, torch.Generator)
                or an int representing the random seed to use for both random states.
        """
        if random_state is None:
            self.random_states = random_state
        elif isinstance(random_state, int):
            self.random_states = (
                np.random.RandomState(seed=random_state),
                torch.Generator().manual_seed(random_state),
            )
        elif (
            isinstance(random_state, tuple)
            and isinstance(random_state[0], np.random.RandomState)
            and isinstance(random_state[1], torch.Generator)
        ):
            self.random_states = random_state
        else:
            raise TypeError(
                f"`random_state` {random_state} expected to be an int or a tuple of "
                "(`np.random.RandomState`, `torch.Generator`)"
            )


class BatchedSynthesizer(BaseSynthesizer):
    def __init__(self, batch_size, **kwargs):
        self._batch_size = batch_size
        super().__init__(**kwargs)

    def get_batch_size(self):
        return self._batch_size

    def set_batch_size(self, b: int):
        warnings.warn("Reset batch_size may caused unintentional behavior.")
        self._batch_size = b



from __future__ import annotations

import hashlib
from functools import cached_property
from typing import Generator

import pandas as pd

from sdgx.data_connectors.base import DataConnector


class CsvConnector(DataConnector):
    """
    Wraps csv file into :ref:`DataConnector`

    Args:
        path (str): Path to csv file
        sep (str, optional): Separator. Defaults to ','.
        header (str, optional): Header. Defaults to 'infer'.
        read_csv_kwargs (dict, optional): kwargs for pd.read_csv, please refer to https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

    Example:

        .. code-block:: python

            from sdgx.data_connectors.csv_connector import CsvConnector
            connector = CsvConnector(
                path="data.csv",
            )
            df = connector.read()


    """

    @cached_property
    def identity(self):
        """
        Identity of the data source is the sha256 of the file
        """
        with open(self.path, "rb") as f:
            return f"csvfile-{hashlib.sha256(f.read()).hexdigest()}"

    def __init__(
        self,
        path,
        sep=",",
        header="infer",
        **read_csv_kwargs,
    ):
        self.path = path
        self.sep = sep
        self.header = header
        self.read_csv_kwargs = read_csv_kwargs

    def _read(self, offset: int = 0, limit: int | None = None) -> pd.DataFrame | None:
        return pd.read_csv(
            self.path,
            sep=self.sep,
            header=self.header,
            skiprows=range(1, offset + 1),  # don't skip header
            nrows=limit,
            **self.read_csv_kwargs,
        )

    def _columns(self) -> list[str]:
        d = pd.read_csv(
            self.path,
            sep=self.sep,
            header=self.header,
            nrows=0,
            **self.read_csv_kwargs,
        ).columns.tolist()
        return d

    def _iter(self, offset: int = 0, chunksize: int = 1000) -> Generator[pd.DataFrame, None, None]:
        if chunksize is None:
            yield self._read(offset=offset)
            return

        for d in pd.read_csv(
            self.path,
            sep=self.sep,
            header=self.header,
            skiprows=range(1, offset + 1),  # don't skip header
            chunksize=chunksize,
            **self.read_csv_kwargs,
        ):
            yield d


from sdgx.data_connectors.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("CsvConnector", CsvConnector)



"""Dataset Generators for numerical transformers."""

from abc import ABC

import numpy as np

from sdgx.models.components.sdv_rdt.performance.datasets.base import (
    BaseDatasetGenerator,
)
from sdgx.models.components.sdv_rdt.performance.datasets.utils import add_nans


class NumericalGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that create numerical data."""

    SDTYPE = "numerical"


class RandomIntegerGenerator(NumericalGenerator):
    """Generator that creates an array of random integers."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        ii32 = np.iinfo(np.int32)
        return np.random.randint(ii32.min, ii32.max, num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 5e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 400.0,
            },
        }


class RandomIntegerNaNsGenerator(NumericalGenerator):
    """Generator that creates an array of random integers with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 4e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 300.0,
            },
        }


class ConstantIntegerGenerator(NumericalGenerator):
    """Generator that creates a constant array with a random integer."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        ii32 = np.iinfo(np.int32)
        constant = np.random.randint(ii32.min, ii32.max)
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 400.0},
            "transform": {"time": 1e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 400.0,
            },
        }


class ConstantIntegerNaNsGenerator(NumericalGenerator):
    """Generator that creates a constant array with a random integer with some nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(ConstantIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 600.0},
            "transform": {"time": 3e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 300.0,
            },
        }


class AlmostConstantIntegerGenerator(NumericalGenerator):
    """Generator that creates an array with 2 only values, one of them repeated."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        ii32 = np.iinfo(np.int32)
        values = np.random.randint(ii32.min, ii32.max, size=2)
        additional_values = np.full(num_rows - 2, values[1])
        array = np.concatenate([values, additional_values])
        np.random.shuffle(array)
        return array

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 1e-05, "memory": 2000.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 2000.0,
            },
        }


class AlmostConstantIntegerNaNsGenerator(NumericalGenerator):
    """Generator that creates an array with 2 only values, one of them repeated, and NaNs."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        ii32 = np.iinfo(np.int32)
        values = np.random.randint(ii32.min, ii32.max, size=2)
        additional_values = np.full(num_rows - 2, values[1]).astype(float)
        array = np.concatenate([values, add_nans(additional_values)])
        np.random.shuffle(array)
        return array

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 3e-05, "memory": 1000.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 1000.0,
            },
        }


class NormalGenerator(NumericalGenerator):
    """Generator that creates an array of normally distributed float values."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.random.normal(size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 1e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 1e-05,
                "memory": 400.0,
            },
        }


class NormalNaNsGenerator(NumericalGenerator):
    """Generator that creates an array of normally distributed float values, with NaNs."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(NormalGenerator.generate(num_rows))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 4e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 300.0,
            },
        }


class BigNormalGenerator(NumericalGenerator):
    """Generator that creates an array of big normally distributed float values."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.random.normal(scale=1e10, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 5e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 5e-05,
                "memory": 400.0,
            },
        }


class BigNormalNaNsGenerator(NumericalGenerator):
    """Generator that creates an array of normally distributed float values, with NaNs."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(BigNormalGenerator.generate(num_rows))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            "fit": {"time": 1e-03, "memory": 2500.0},
            "transform": {"time": 3e-05, "memory": 400.0},
            "reverse_transform": {
                "time": 2e-05,
                "memory": 300.0,
            },
        }



"""DataSampler module."""

import numpy as np


class DataSampler(object):
    """DataSampler samples the conditional vector and corresponding data for CTGAN."""

    def __init__(self, data, output_info, log_frequency):
        self._data = data

        def is_discrete_column(column_info):
            return len(column_info) == 1 and column_info[0].activation_fn == "softmax"

        n_discrete_columns = sum(
            [1 for column_info in output_info if is_discrete_column(column_info)]
        )

        self._discrete_column_matrix_st = np.zeros(n_discrete_columns, dtype="int32")

        # Store the row id for each category in each discrete column.
        # For example _rid_by_cat_cols[a][b] is a list of all rows with the
        # a-th discrete column equal value b.
        self._rid_by_cat_cols = []

        # Compute _rid_by_cat_cols
        st = 0
        for column_info in output_info:
            if is_discrete_column(column_info):
                span_info = column_info[0]
                ed = st + span_info.dim

                rid_by_cat = []
                for j in range(span_info.dim):
                    rid_by_cat.append(np.nonzero(data[:, st + j])[0])
                self._rid_by_cat_cols.append(rid_by_cat)
                st = ed
            else:
                st += sum([span_info.dim for span_info in column_info])
        assert st == data.shape[1]

        # Prepare an interval matrix for efficiently sample conditional vector
        max_category = max(
            [column_info[0].dim for column_info in output_info if is_discrete_column(column_info)],
            default=0,
        )

        self._discrete_column_cond_st = np.zeros(n_discrete_columns, dtype="int32")
        self._discrete_column_n_category = np.zeros(n_discrete_columns, dtype="int32")
        self._discrete_column_category_prob = np.zeros((n_discrete_columns, max_category))
        self._n_discrete_columns = n_discrete_columns
        self._n_categories = sum(
            [column_info[0].dim for column_info in output_info if is_discrete_column(column_info)]
        )

        st = 0
        current_id = 0
        current_cond_st = 0
        for column_info in output_info:
            if is_discrete_column(column_info):
                span_info = column_info[0]
                ed = st + span_info.dim
                category_freq = np.sum(data[:, st:ed], axis=0)
                if log_frequency:
                    category_freq = np.log(category_freq + 1)
                category_prob = category_freq / np.sum(category_freq)
                self._discrete_column_category_prob[current_id, : span_info.dim] = category_prob
                self._discrete_column_cond_st[current_id] = current_cond_st
                self._discrete_column_n_category[current_id] = span_info.dim
                current_cond_st += span_info.dim
                current_id += 1
                st = ed
            else:
                st += sum([span_info.dim for span_info in column_info])

    def _random_choice_prob_index(self, discrete_column_id):
        probs = self._discrete_column_category_prob[discrete_column_id]
        r = np.expand_dims(np.random.rand(probs.shape[0]), axis=1)
        return (probs.cumsum(axis=1) > r).argmax(axis=1)

    def sample_condvec(self, batch):
        """Generate the conditional vector for training.

        Returns:
            cond (batch x #categories):
                The conditional vector.
            mask (batch x #discrete columns):
                A one-hot vector indicating the selected discrete column.
            discrete column id (batch):
                Integer representation of mask.
            category_id_in_col (batch):
                Selected category in the selected discrete column.
        """
        if self._n_discrete_columns == 0:
            return None

        discrete_column_id = np.random.choice(np.arange(self._n_discrete_columns), batch)

        cond = np.zeros((batch, self._n_categories), dtype="float32")
        mask = np.zeros((batch, self._n_discrete_columns), dtype="float32")
        mask[np.arange(batch), discrete_column_id] = 1
        category_id_in_col = self._random_choice_prob_index(discrete_column_id)
        category_id = self._discrete_column_cond_st[discrete_column_id] + category_id_in_col
        cond[np.arange(batch), category_id] = 1

        return cond, mask, discrete_column_id, category_id_in_col

    def sample_original_condvec(self, batch):
        """Generate the conditional vector for generation use original frequency."""
        if self._n_discrete_columns == 0:
            return None

        cond = np.zeros((batch, self._n_categories), dtype="float32")

        for i in range(batch):
            row_idx = np.random.randint(0, len(self._data))
            col_idx = np.random.randint(0, self._n_discrete_columns)
            matrix_st = self._discrete_column_matrix_st[col_idx]
            matrix_ed = matrix_st + self._discrete_column_n_category[col_idx]
            pick = np.argmax(self._data[row_idx, matrix_st:matrix_ed])
            cond[i, pick + self._discrete_column_cond_st[col_idx]] = 1

        return cond

    def sample_data(self, n, col, opt):
        """Sample data from original training data satisfying the sampled conditional vector.

        Returns:
            n rows of matrix data.
        """
        if col is None:
            idx = np.random.randint(len(self._data), size=n)
            return self._data[idx]

        idx = []
        for c, o in zip(col, opt):
            idx.append(np.random.choice(self._rid_by_cat_cols[c][o]))

        return self._data[idx]

    def dim_cond_vec(self):
        """Return the total number of categories."""
        return self._n_categories

    def generate_cond_from_condition_column_info(self, condition_info, batch):
        """Generate the condition vector."""
        vec = np.zeros((batch, self._n_categories), dtype="float32")
        id_ = self._discrete_column_matrix_st[condition_info["discrete_column_id"]]
        id_ += condition_info["value_id"]
        vec[:, id_] = 1
        return vec



from __future__ import annotations

from typing import Dict, List, Set

import numpy as np
import pandas as pd

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class SpecificCombinationTransformer(Transformer):
    """
    A transformer used to handle specific combinations of columns in tabular data.

    The relationships between columns can be quite complex. Currently, we introduced `FixedCombinationTransformer`
    is not capable of comprehensive automatic detection. This transformer allows users to manually specify the
    mapping relationships between columns, specifically for multiple corresponding relationships. Users can define
    multiple groups, with each group supporting multiple columns. The transformer will record the combination values
    of each column, and in the `reverse_convert()`, it will restore any mismatched combinations from the recorded
    relationships.

    For example:

    | Category A | Category B | Category C | Category D | Category E |
    | :--------: | :--------: | :--------: | :--------: | :--------: |
    |     A1     |     B1     |     C1     |     D1     |     E1     |
    |     A1     |     B1     |     C2     |     D2     |     E2     |
    |     A2     |     B2     |     C1     |     D1     |     E3     |

    Here user can specific combination like (Category A, Category B), (Category C, Category D, Category E).

    For now, the `specific_combinations` passing by `Metadata`

    """

    column_groups: List[Set[str]]
    """
    Define a list where each element is a set containing string type column names
    """

    mappings: Dict[frozenset, pd.DataFrame]
    """
    Define a dictionary variable `mappings` where the keys are frozensets and the values are pandas DataFrame objects
    """

    specified: bool
    """
    Define a boolean that flag if user specified the combination, if true, that handle the `specific_combinations`
    """

    def __init__(self):
        self.column_groups: List[Set[str]] = []
        self.mappings: Dict[frozenset, pd.DataFrame] = {}
        self.specified = False

    def fit(self, metadata: Metadata | None = None, tabular_data: DataLoader | pd.DataFrame = None):
        """
        Study the combination relationships and value mapping of columns.

        Args:
            metadata: Metadata containing information about specific column combinations.
            tabular_data: The tabular data to be fitted, can be a DataLoader object or a pandas DataFrame.
        """
        specific_combinations = metadata.get("specific_combinations")
        if specific_combinations is None or len(specific_combinations) == 0:
            logger.info(
                "Fit data using SpecificCombinationTransformer(No specified)... Finished (No action)."
            )
            self.fitted = True
            return

        # Create a mapping relationship for each group of columns
        df = tabular_data
        self.column_groups = [set(cols) for cols in specific_combinations]
        for group in self.column_groups:
            group_df = df[list(group)].drop_duplicates()
            self.mappings[frozenset(group)] = group_df

        self.fitted = True
        self.specified = True
        logger.info("SpecificCombinationTransformer Fitted.")

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert the raw data based on the learned mapping relationships.

        Args:
           raw_data: The raw data to be converted.

        Returns:
           The converted data.
        """
        if not self.specified:
            logger.info(
                "Converting data using SpecificCombinationTransformer(No specified)... Finished (No action)."
            )
            return super().convert(raw_data)

        logger.info("SpecificCombinationTransformer convert doing nothing...")
        return super().convert(raw_data)

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse convert the processed data to ensure it conforms to the original format.

        Args:
            processed_data: The processed data to be reverse converted.

        Returns:
            The reverse converted data.
        """
        if not self.specified:
            logger.info(
                "Reverse converting data using SpecificCombinationTransformer(No specified)... Finished (No action)."
            )
            return processed_data

        result_df = processed_data.copy()
        n_rows = len(result_df)

        # For each column-mapping groups, replace with random choice
        # Here we random_indices for len(processed_data) from column-mapping and replaced processed_data.
        for group in self.column_groups:
            group_mapping = self.mappings[frozenset(group)]
            group_cols = list(group)
            random_indices = np.random.choice(len(group_mapping), size=n_rows)
            random_mappings = group_mapping.iloc[random_indices]
            result_df[group_cols] = random_mappings[group_cols].values

        return result_df


@hookimpl
def register(manager):
    manager.register("SpecificCombinationTransformer", SpecificCombinationTransformer)



import pandas as pd
import pytest

from sdgx.data_models.inspectors.discrete import DiscreteInspector


@pytest.fixture
def inspector():
    yield DiscreteInspector()


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


def test_inspector(inspector: DiscreteInspector, raw_data):
    inspector.fit(raw_data)
    assert inspector.ready
    assert inspector.discrete_columns
    assert sorted(inspector.inspect()["discrete_columns"]) == sorted(
        [
            "workclass",
            "education",
            "marital-status",
            "occupation",
            "relationship",
            "race",
            "gender",
            "native-country",
            "income",
        ]
    )
    assert inspector.inspect_level == 10


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pytest

from sdgx.models.manager import ModelManager


@pytest.fixture
def manager():
    yield ModelManager()


@pytest.mark.parametrize(
    "supported_model",
    [
        "ctgan",
    ],
)
def test_manager(supported_model, manager: ModelManager):
    assert manager._normalize_name(supported_model) in manager.registed_models
    manager.init_model(supported_model, epochs=1)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from pathlib import Path

import pandas as pd

_HERE = Path(__file__).parent

dataset_csv = (_HERE / "dataset/benchmark.csv").expanduser().resolve()
df = pd.read_csv(dataset_csv)

discrete_columns = [s for s in df.columns if s.startswith("string")]


from ctgan import CTGAN

ctgan = CTGAN(epochs=1, cuda=False)
ctgan.fit(df, discrete_columns)



from __future__ import annotations

from typing import Generator

import pandas as pd
import pytest

from sdgx.cachers.base import Cacher, NoCache
from sdgx.cachers.disk_cache import DiskCache
from sdgx.data_connectors.csv_connector import CsvConnector


@pytest.fixture
def csv_file(tmp_path):
    csv = tmp_path / "test.csv"
    csv.write_text("index,a,b,c\n0,1,2,3\n1,4,5,6")
    yield csv
    csv.unlink()


class MockCsvConnector(CsvConnector):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._readed = False

    @property
    def is_readed(self):
        return self._readed

    def reset(self):
        self._readed = False

    def _read(self, offset=0, limit=0) -> pd.DataFrame:
        self._readed = True
        try:
            return super()._read(offset, limit)
        except Exception:
            self._readed = False
            raise

    def _columns(self) -> list[str]:
        self._readed = True
        try:
            return super()._columns()
        except Exception:
            self._readed = False
            raise

    def iter(self, offset=0, chunksize=0) -> Generator[pd.DataFrame, None, None]:
        self._readed = True
        try:
            return super().iter(offset, chunksize)
        except Exception:
            self._readed = False
            raise


@pytest.fixture
def data_connector(csv_file):
    yield MockCsvConnector(
        path=csv_file,
    )


@pytest.mark.parametrize(
    "cacher_cls",
    [
        NoCache,
        DiskCache,
    ],
)
@pytest.mark.parametrize("blocksize", [1])
@pytest.mark.parametrize("chunksize", [1])
def test_cacher(cacher_cls, cacher_kwargs, blocksize, chunksize, data_connector):
    cacher: Cacher = cacher_cls(blocksize=blocksize, **cacher_kwargs)
    for d in cacher.iter(chunksize, data_connector):
        assert len(d) == chunksize
    if isinstance(cacher, NoCache):
        assert not cacher.is_cached(0)
    else:
        assert cacher.is_cached(0)

        data_connector.reset()
        cacher.load(0, chunksize, data_connector)
        assert not data_connector.is_readed
    assert not cacher.load_all(data_connector).empty


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



"""UniformUnivariate module."""

import numpy as np
from scipy.stats import uniform

from sdgx.models.components.sdv_copulas.univariate.base import (
    BoundedType,
    ParametricType,
    ScipyModel,
)


class UniformUnivariate(ScipyModel):
    """Uniform univariate model."""

    PARAMETRIC = ParametricType.PARAMETRIC
    BOUNDED = BoundedType.BOUNDED

    MODEL_CLASS = uniform

    def _fit_constant(self, X):
        self._params = {"loc": np.min(X), "scale": np.max(X) - np.min(X)}

    def _fit(self, X):
        self._params = {"loc": np.min(X), "scale": np.max(X) - np.min(X)}

    def _is_constant(self):
        return self._params["scale"] == 0

    def _extract_constant(self):
        return self._params["loc"]



import pytest

from sdgx.models.manager import ModelManager


@pytest.fixture
def manager():
    yield ModelManager()


def test_registed_model(manager: ModelManager):
    assert manager._normalize_name("DummyModel") in manager.registed_models


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from typing import Any

import pandas as pd
from pandas._libs.tslibs.parsing import DateParseError

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class BoolInspector(Inspector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bool_columns: set[str] = set()

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """Fit the inspector.

        Gets the list of discrete columns from the raw data.

        Args:
            raw_data (pd.DataFrame): Raw data
        """
        self.bool_columns = set()
        self.bool_columns = self.bool_columns.union(
            set(raw_data.infer_objects().select_dtypes(include=["bool"]).columns)
        )

        self.ready = True

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""

        return {"bool_columns": list(self.bool_columns)}


@hookimpl
def register(manager):
    manager.register("BoolInspector", BoolInspector)



from __future__ import annotations

import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from sdgx.data_loader import DataLoader
from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.models.components.optimize.ndarray_loader import NDArrayLoader
from sdgx.utils import logger


class DiscreteTransformer(Transformer):
    """
    A transformer class for handling discrete values in the input data.

    This class uses one-hot encoding to convert discrete values into a format that can be used by machine learning models.

    Attributes:
        discrete_columns (list): A list of column names that are of discrete type.
        one_hot_warning_cnt (int): The warning count for one-hot encoding. If the number of new columns after one-hot encoding exceeds this count, a warning message will be issued.
        one_hot_encoders (dict): A dictionary that stores the OneHotEncoder objects for each discrete column. The keys are the column names, and the values are the corresponding OneHotEncoder objects.
        one_hot_column_names (dict): A dictionary that stores the new column names after one-hot encoding for each discrete column. The keys are the column names, and the values are lists of new column names.
        onehot_encoder_handle_unknown (str): The parameter to handle unknown categories in the OneHotEncoder. If set to 'ignore', new categories will be ignored. If set to 'error', an error will be raised when new categories are encountered.

    Methods:
        fit(metadata: Metadata, tabular_data: DataLoader | pd.DataFrame): Fit the transformer to the input data.
        _fit_column(column_name: str, column_data: pd.DataFrame): Fit a single discrete column.
        convert(raw_data: pd.DataFrame) -> pd.DataFrame: Convert the input data using one-hot encoding.
        reverse_convert(processed_data: pd.DataFrame) -> pd.DataFrame: Reverse the one-hot encoding process to get the original data.
    """

    discrete_columns: list
    """
    Record which columns are of discrete type.
    """

    one_hot_warning_cnt: int
    """
    The warning count for one-hot encoding.
    If the number of new columns after one-hot encoding exceeds this count, a warning message will be issued.
    """

    one_hot_encoders: dict
    """
    A dictionary that stores the OneHotEncoder objects for each discrete column.
    The keys are the column names, and the values are the corresponding OneHotEncoder objects.
    """

    one_hot_column_names: dict
    """
    A dictionary that stores the new column names after one-hot encoding for each discrete column.
    The keys are the column names, and the values are lists of new column names.
    """

    onehot_encoder_handle_unknown: str
    """
    The parameter to handle unknown categories in the OneHotEncoder.
    If set to 'ignore', new categories will be ignored.
    If set to 'error', an error will be raised when new categories are encountered.
    """

    def __init__(self):
        self.discrete_columns = []
        self.one_hot_warning_cnt = 512
        self.one_hot_encoders = {}
        self.one_hot_column_names = {}
        self.onehot_encoder_handle_unknown = "ignore"

    def fit(self, metadata: Metadata, tabular_data: DataLoader | pd.DataFrame):
        """
        Fit method for the DiscreteTransformer.
        """

        logger.info("Fitting using DiscreteTransformer...")

        self.discrete_columns = metadata.get("discrete_columns")

        # remove datetime columns from discrete columns
        # because datetime columns are converted into timestamps
        datetime_columns = metadata.get("datetime_columns")

        # no discrete columns
        if len(self.discrete_columns) == 0:
            logger.info("Fitting using DiscreteTransformer... Finished (No Columns).")
            return

        # fit each columns
        for each_datgetime_col in datetime_columns:
            if each_datgetime_col in self.discrete_columns:
                self.discrete_columns.remove(each_datgetime_col)
                logger.info(f"Datetime column {each_datgetime_col} removed from discrete column.")

        # then, there are >= 1 discrete colums
        for each_col in self.discrete_columns:
            # fit each column
            self._fit_column(each_col, tabular_data[[each_col]])

        logger.info("Fitting using DiscreteTransformer... Finished.")
        self.fitted = True

        return

    def _fit_column(self, column_name: str, column_data: pd.DataFrame):
        """
        Fit every discrete column in `_fit_column`.

        Args:
            - column_data (pd.DataFrame): A dataframe containing a column.
            - column_name: str: column name.
        """

        self.one_hot_encoders[column_name] = OneHotEncoder(
            handle_unknown=self.onehot_encoder_handle_unknown, sparse_output=False
        )
        # fit the column data
        self.one_hot_encoders[column_name].fit(column_data)

        logger.debug(f"Discrete column {column_name} fitted.")

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Convert method to handle discrete values in the input data.
        """

        logger.info("Converting data using DiscreteTransformer...")

        # transform every discrete column into one-hot encoded columns
        if len(self.discrete_columns) == 0:
            logger.info("Converting data using DiscreteTransformer... Finished (No column).")
            return

        processed_data = raw_data.copy()

        for each_col in self.discrete_columns:
            # 1- transform each column
            new_onehot_columns = self.one_hot_encoders[each_col].transform(raw_data[[each_col]])
            new_onehot_column_names = self.one_hot_encoders[each_col].get_feature_names_out()
            self.one_hot_column_names[each_col] = new_onehot_column_names

            # logger warning if too many columns
            if len(new_onehot_column_names) > self.one_hot_warning_cnt:
                logger.warning(
                    f"Column {each_col} has too many discrete values ({len(new_onehot_column_names)} values), may consider as a continous column?"
                )

            # 2- add new_onehot_column_set into the original dataframe, record the column name ?
            processed_data = self.attach_columns(
                processed_data, pd.DataFrame(new_onehot_columns, columns=new_onehot_column_names)
            )

            logger.debug(f"Column {each_col} converted.")

        logger.info(f"Processed data shape: {processed_data.shape}.")

        logger.info("Converting data using DiscreteTransformer... Finished.")

        processed_data = self.remove_columns(processed_data, self.discrete_columns)

        return processed_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverse_convert method for the transformer.

        Args:
            - processed_data (pd.DataFrame): A dataframe containing onehot encoded columns.

        Returns:
            - pd.DataFrame: inverse transformed processed data.
        """

        reversed_data = processed_data.copy()

        # for each discrete col
        for each_col in self.discrete_columns:
            # 1- get one-hot column sets from processed data
            one_hot_column_set = processed_data[self.one_hot_column_names[each_col]]
            # 2- inverse convert using ohe
            res_column_data = self.one_hot_encoders[each_col].inverse_transform(
                pd.DataFrame(one_hot_column_set, columns=self.one_hot_column_names[each_col])
            )
            # 3- put original column back to reversed_data
            reversed_data = self.attach_columns(
                reversed_data, pd.DataFrame(res_column_data, columns=[each_col])
            )
            reversed_data = self.remove_columns(reversed_data, self.one_hot_column_names[each_col])

        logger.info("Data inverse-converted by DiscreteTransformer.")

        return reversed_data

    pass


"""
@hookimpl
def register(manager):
    manager.register("DiscreteTransformer", DiscreteTransformer)
"""



"""Functions to profile performance of RDT Transformers."""

# pylint: disable=W0212

import multiprocessing as mp
import timeit
import tracemalloc
from copy import deepcopy

import pandas as pd


def _profile_time(transformer, method_name, dataset, column=None, iterations=10, copy=False):
    total_time = 0
    for _ in range(iterations):
        if copy:
            transformer_copy = deepcopy(transformer)
            method = getattr(transformer_copy, method_name)

        else:
            method = getattr(transformer, method_name)

        start_time = timeit.default_timer()
        if column:
            method(dataset, column)
        else:
            method(dataset)
        total_time += timeit.default_timer() - start_time

    return total_time / iterations


def _set_memory_for_method(method, dataset, column, peak_memory):
    tracemalloc.start()
    if column:
        method(dataset, column)
    else:
        method(dataset)

    peak_memory.value = tracemalloc.get_traced_memory()[1]
    tracemalloc.stop()
    tracemalloc.clear_traces()


def _profile_memory(method, dataset, column=None):
    ctx = mp.get_context("spawn")
    peak_memory = ctx.Value("i", 0)
    profiling_process = ctx.Process(
        target=_set_memory_for_method, args=(method, dataset, column, peak_memory)
    )
    profiling_process.start()
    profiling_process.join()
    return peak_memory.value


def profile_transformer(transformer, dataset_generator, transform_size, fit_size=None):
    """Profile a Transformer on a dataset.

    This function will get the total time and peak memory
    for the ``fit``, ``transform`` and ``reverse_transform``
    methods of the provided transformer against the provided
    dataset.

    Args:
        transformer (Transformer):
            Transformer instance.
        dataset_generator (DatasetGenerator):
            DatasetGenerator instance.
        transform_size (int):
            Number of rows to generate for ``transform`` and ``reverse_transform``.
        fit_size (int or None):
            Number of rows to generate for ``fit``. If None, use ``transform_size``.

    Returns:
        pandas.Series:
            Series containing the time and memory taken by ``fit``, ``transform``,
            and ``reverse_transform`` for the transformer.
    """
    fit_size = fit_size or transform_size
    fit_dataset = pd.DataFrame({"test": dataset_generator.generate(fit_size)})
    replace = transform_size > fit_size
    transform_dataset = fit_dataset.sample(transform_size, replace=replace)

    fit_time = _profile_time(transformer, "fit", fit_dataset, column="test", copy=True)
    fit_memory = _profile_memory(transformer.fit, fit_dataset, column="test")
    transformer.fit(fit_dataset, "test")

    transform_time = _profile_time(transformer, "transform", transform_dataset)
    transform_memory = _profile_memory(transformer.transform, transform_dataset)

    reverse_dataset = transformer.transform(transform_dataset)
    reverse_time = _profile_time(transformer, "reverse_transform", reverse_dataset)
    reverse_memory = _profile_memory(transformer.reverse_transform, reverse_dataset)

    return pd.Series(
        {
            "Fit Time": fit_time,
            "Fit Memory": fit_memory,
            "Transform Time": transform_time,
            "Transform Memory": transform_memory,
            "Reverse Transform Time": reverse_time,
            "Reverse Transform Memory": reverse_memory,
        }
    )



import datetime
import random
import string

import pandas as pd
import pytest
from faker import Faker

fake = Faker(locale="zh_CN")
fake_en = Faker(["en_US"])

from sdgx.data_models.inspectors.personal import (
    ChinaMainlandAddressInspector,
    ChinaMainlandIDInspector,
    ChinaMainlandMobilePhoneInspector,
    ChinaMainlandPostCode,
    ChinaMainlandUnifiedSocialCreditCode,
    ChineseCompanyNameInspector,
    ChineseNameInspector,
    EmailInspector,
    EnglishNameInspector,
)


def generate_uniform_credit_code():
    # generate china mainland 统一社会信用代码 for test
    def generate():
        code = ""
        code += random.choice(string.digits + "AHJNPQRTUWXY")
        code += random.choice(string.digits + "AHJNPQRTUWXY")
        code += "".join(random.choices(string.digits, k=6))
        code += "".join(random.choices(string.digits, k=9))
        code += random.choice(string.digits + "AHJNPQRTUWXY")
        return code

    code = generate()
    return code


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


@pytest.fixture
def chn_personal_test_df():
    row_cnt = 1000
    today = datetime.datetime.today()
    X = []
    header = [
        "ssn_sfz",
        "chn_name",
        "eng_name",
        "gender",
        "birth_date",
        "age",
        "email",
        "mobile_phone_no",
        "chn_address",
        "postcode",
        "job",
        "company_name",
        "uscc",
    ]
    for _ in range(row_cnt):
        each_gender = random.choice(["male", "female"])
        if each_gender == "male":
            each_name = fake.last_name() + fake.name_male()
        else:
            each_name = fake.last_name() + fake.name_female()
        each_eng_name = fake_en.name()
        each_birth_date = fake.date()
        each_age = today.year - int(each_birth_date[:4])
        each_email = fake.email()
        each_phone = fake.phone_number()
        each_sfz = fake.ssn()
        each_address = fake.address()
        each_job = fake.job()
        each_corp = fake.company()
        each_postcode = fake.postcode()
        each_uscc = generate_uniform_credit_code()

        each_x = [
            each_sfz,
            each_name,
            each_eng_name,
            each_gender,
            each_birth_date,
            each_age,
            each_email,
            each_phone,
            each_address,
            each_postcode,
            each_job,
            each_corp,
            each_uscc,
        ]

        X.append(each_x)

    yield pd.DataFrame(X, columns=header)


# Email
def test_email_inspector_demo_data(raw_data):
    inspector_Email = EmailInspector()
    inspector_Email.fit(raw_data)
    assert not inspector_Email.regex_columns
    assert sorted(inspector_Email.inspect()["email_columns"]) == sorted([])
    assert inspector_Email.inspect_level == 30
    assert inspector_Email.pii is True


def test_email_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_Email = EmailInspector()
    inspector_Email.fit(chn_personal_test_df)
    assert inspector_Email.regex_columns
    assert sorted(inspector_Email.inspect()["email_columns"]) == sorted(["email"])
    assert inspector_Email.inspect_level == 30
    assert inspector_Email.pii is True


# Phone No
def test_chn_phone_inspector_demo_data(raw_data):
    inspector_Phone = ChinaMainlandMobilePhoneInspector()
    inspector_Phone.fit(raw_data)
    assert not inspector_Phone.regex_columns
    assert sorted(inspector_Phone.inspect()["china_mainland_mobile_phone_columns"]) == sorted([])
    assert inspector_Phone.inspect_level == 30
    assert inspector_Phone.pii is True


def test_chn_phone_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_Phone = ChinaMainlandMobilePhoneInspector()
    inspector_Phone.fit(chn_personal_test_df)
    assert inspector_Phone.regex_columns
    assert sorted(inspector_Phone.inspect()["china_mainland_mobile_phone_columns"]) == sorted(
        ["mobile_phone_no"]
    )
    assert inspector_Phone.inspect_level == 30
    assert inspector_Phone.pii is True


# China Mainland ID / 居民身份证
def test_chn_ID_inspector_demo_data(raw_data):
    inspector_ID = ChinaMainlandIDInspector()
    inspector_ID.fit(raw_data)
    assert not inspector_ID.regex_columns
    assert sorted(inspector_ID.inspect()["china_mainland_id_columns"]) == sorted([])
    assert inspector_ID.inspect_level == 30
    assert inspector_ID.pii is True


def test_chn_ID_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_ID = ChinaMainlandIDInspector()
    inspector_ID.fit(chn_personal_test_df)
    assert inspector_ID.regex_columns
    assert sorted(inspector_ID.inspect()["china_mainland_id_columns"]) == sorted(["ssn_sfz"])
    assert inspector_ID.inspect_level == 30
    assert inspector_ID.pii is True


# PostCode
def test_chn_postcode_inspector_demo_data(raw_data):
    inspector_PostCode = ChinaMainlandPostCode()
    inspector_PostCode.fit(raw_data)
    assert not inspector_PostCode.regex_columns
    assert sorted(inspector_PostCode.inspect()["china_mainland_postcode_columns"]) == sorted([])
    assert inspector_PostCode.inspect_level == 20
    assert inspector_PostCode.pii is False


def test_chn_postcode_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_PostCode = ChinaMainlandPostCode()
    inspector_PostCode.fit(chn_personal_test_df)
    assert inspector_PostCode.regex_columns
    assert sorted(inspector_PostCode.inspect()["china_mainland_postcode_columns"]) == sorted(
        ["postcode"]
    )
    assert inspector_PostCode.inspect_level == 20
    assert inspector_PostCode.pii is False


# 统一社会信用代码
def test_chn_uscc_inspector_demo_data(raw_data):
    inspector_USCC = ChinaMainlandUnifiedSocialCreditCode()
    inspector_USCC.fit(raw_data)
    assert not inspector_USCC.regex_columns
    assert sorted(inspector_USCC.inspect()["unified_social_credit_code_columns"]) == sorted([])
    assert inspector_USCC.inspect_level == 30
    assert inspector_USCC.pii is True


def test_chn_uscc_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_USCC = ChinaMainlandUnifiedSocialCreditCode()
    inspector_USCC.fit(chn_personal_test_df)
    assert inspector_USCC.regex_columns
    assert sorted(inspector_USCC.inspect()["unified_social_credit_code_columns"]) == sorted(
        ["uscc"]
    )
    assert inspector_USCC.inspect_level == 30
    assert inspector_USCC.pii is True


# CHN address
def test_chn_address_inspector_demo_data(raw_data):
    inspector_CHN_Address = ChinaMainlandAddressInspector()
    inspector_CHN_Address.fit(raw_data)
    assert not inspector_CHN_Address.regex_columns
    assert sorted(inspector_CHN_Address.inspect()["china_mainland_address_columns"]) == sorted([])
    assert inspector_CHN_Address.inspect_level == 30
    assert inspector_CHN_Address.pii is True


def test_chn_address_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_CHN_Address = ChinaMainlandAddressInspector()
    inspector_CHN_Address.fit(chn_personal_test_df)
    assert inspector_CHN_Address.regex_columns
    assert sorted(inspector_CHN_Address.inspect()["china_mainland_address_columns"]) == sorted(
        ["chn_address"]
    )
    assert inspector_CHN_Address.inspect_level == 30
    assert inspector_CHN_Address.pii is True


# CHN name
def test_chn_name_inspector_demo_data(raw_data):
    inspector_CHN_name = ChineseNameInspector()
    inspector_CHN_name.fit(raw_data)
    assert not inspector_CHN_name.regex_columns
    assert sorted(inspector_CHN_name.inspect()["chinese_name_columns"]) == sorted([])
    assert inspector_CHN_name.inspect_level == 40
    assert inspector_CHN_name.pii is True


def test_chn_name_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_CHN_name = ChineseNameInspector()
    inspector_CHN_name.fit(chn_personal_test_df)
    assert inspector_CHN_name.regex_columns
    assert sorted(inspector_CHN_name.inspect()["chinese_name_columns"]) == sorted(["chn_name"])
    assert inspector_CHN_name.inspect_level == 40
    assert inspector_CHN_name.pii is True


# English name
def test_eng_name_inspector_demo_data(raw_data):
    inspector_ENG_name = EnglishNameInspector()
    inspector_ENG_name.fit(raw_data)
    assert not inspector_ENG_name.regex_columns
    assert sorted(inspector_ENG_name.inspect()["english_name_columns"]) == sorted([])
    assert inspector_ENG_name.inspect_level == 40
    assert inspector_ENG_name.pii is True


def test_eng_name_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_ENG_name = EnglishNameInspector()
    inspector_ENG_name.fit(chn_personal_test_df)
    assert inspector_ENG_name.regex_columns
    assert sorted(inspector_ENG_name.inspect()["english_name_columns"]) == sorted(["eng_name"])
    assert inspector_ENG_name.inspect_level == 40
    assert inspector_ENG_name.pii is True


# Chinese Company Name
def test_chn_company_inspector_demo_data(raw_data):
    inspector_PostCode = ChineseCompanyNameInspector()
    inspector_PostCode.fit(raw_data)
    assert not inspector_PostCode.regex_columns
    assert sorted(inspector_PostCode.inspect()["chinese_company_name_columns"]) == sorted([])
    assert inspector_PostCode.inspect_level == 40
    assert inspector_PostCode.pii is False


def test_chn_company_inspector_generated_data(chn_personal_test_df: pd.DataFrame):
    inspector_PostCode = ChineseCompanyNameInspector()
    inspector_PostCode.fit(chn_personal_test_df)
    assert inspector_PostCode.regex_columns
    assert sorted(inspector_PostCode.inspect()["chinese_company_name_columns"]) == sorted(
        ["company_name"]
    )
    assert inspector_PostCode.inspect_level == 40
    assert inspector_PostCode.pii is False


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import datetime

import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.inspectors.regex import RegexInspector
from sdgx.exceptions import InspectorInitError


@pytest.fixture
def int_inspector():
    yield RegexInspector(pattern="^[0-9]*$", data_type_name="int")


@pytest.fixture
def empty_inspector():
    yield RegexInspector(pattern="^$", data_type_name="empty_columns", match_percentage=0.88)


@pytest.fixture
def raw_data(demo_single_table_path):
    yield pd.read_csv(demo_single_table_path)


def test_int_regex_inspector_demo_data(int_inspector: RegexInspector, raw_data: pd.DataFrame):
    int_inspector.fit(raw_data)
    assert int_inspector.ready
    # should not be empty set
    assert int_inspector.regex_columns
    assert sorted(int_inspector.inspect()["int_columns"]) == sorted(
        ["age", "capital-gain", "capital-loss", "educational-num", "fnlwgt", "hours-per-week"]
    )
    assert int_inspector.inspect_level == 10


def test_empty_regex_inspector_demo_data(empty_inspector: RegexInspector, raw_data: pd.DataFrame):
    empty_inspector.fit(raw_data)
    assert empty_inspector.ready
    # should be empty set
    assert not empty_inspector.regex_columns
    assert sorted(empty_inspector.inspect()["empty_columns"]) == sorted([])
    assert empty_inspector.inspect_level == 10


def test_match_rate_property(empty_inspector: RegexInspector):
    assert empty_inspector.match_percentage == 0.88
    empty_inspector.match_percentage = 0.7
    empty_inspector.match_percentage = 1

    try:
        empty_inspector.match_percentage = 1.2
    except Exception as e:
        assert type(e) == InspectorInitError

    try:
        empty_inspector.match_percentage = 0.5
    except Exception as e:
        assert type(e) == InspectorInitError
    pass


def test_parameter_missing_case():
    # init only with pattern
    only_pattern_inspector = RegexInspector(pattern="^[0-9]*$")
    assert only_pattern_inspector.data_type_name == "regex_^[0-9]*$_columns"
    # init without pattern
    has_error = False
    try:
        miss_pattern_inspector = RegexInspector(data_type_name="xx")
    except Exception as e:
        has_error = True
        assert type(e) == InspectorInitError
    assert has_error is True
    # init without data type name and pattern
    has_error = False
    try:
        dtype_pattern_inspector = RegexInspector()
    except Exception as e:
        has_error = True
        assert type(e) == InspectorInitError
    assert has_error is True


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from typing import Any

import pandas as pd
from faker import Faker

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.generators.pii import PIIGenerator

fake = Faker()


class EmailGenerator(PIIGenerator):
    """
    A class for generating and reversing the conversion of email addresses in a pd.DataFrame.

    This class is a subclass of `PIIGenerator` and is designed to handle the conversion and
    reversal of email addresses in a pd.DataFrame. It uses the `email_columns_list` to identify
    which columns in the pd.DataFrame contain email addresses.

    Attributes:
        email_columns_list (list): A list of column names in the pd.DataFrame that contain email addresses.

    Methods:
        fit(metadata: Metadata | None = None): Fits the generator to the metadata.
        convert(raw_data: pd.DataFrame) -> pd.DataFrame: Converts the email addresses in the pd.DataFrame.
        reverse_convert(processed_data: pd.DataFrame) -> pd.DataFrame: Reverses the conversion of the email addresses in the pd.DataFrame.
    """

    email_columns_list: list

    fitted: bool

    def __init__(self):
        self.email_columns_list = []
        self.fitted = False

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):

        self.email_columns_list = list(metadata.get("email_columns"))

        self.fitted = True

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        if not self.email_columns_list:
            return raw_data
        processed_data = raw_data
        # remove every email column from the dataframe
        for each_col in self.email_columns_list:
            processed_data = self.remove_columns(processed_data, each_col)
        return processed_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        if not self.email_columns_list:
            return processed_data
        df_length = processed_data.shape[0]
        for each_col_name in self.email_columns_list:
            each_email_col = [fake.ascii_company_email() for _ in range(df_length)]
            each_email_df = pd.DataFrame({each_col_name: each_email_col})
            processed_data = self.attach_columns(processed_data, each_email_df)

        return processed_data


@hookimpl
def register(manager):
    manager.register("EmailGenerator", EmailGenerator)



"""Base class for all the Dataset Generators."""

from abc import ABC, abstractmethod


class BaseDatasetGenerator(ABC):
    """Parent class for all the Dataset Generators."""

    SDTYPE = None

    @staticmethod
    @abstractmethod
    def generate(num_rows):
        """Return array of data. This method serves as a template for dataset generators.

        Args:
            num_rows (int):
                Number of rows to generate.

        Returns:
            numpy.ndarray of size ``num_rows``
        """
        raise NotImplementedError()

    @classmethod
    def get_subclasses(cls):
        """Recursively find subclasses of this Baseline.

        Returns:
            list:
                List of all subclasses of this class.
        """
        subclasses = []
        for subclass in cls.__subclasses__():
            if ABC not in subclass.__bases__:
                subclasses.append(subclass)

            subclasses += subclass.get_subclasses()

        return subclasses

    @staticmethod
    @abstractmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        raise NotImplementedError()



from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class EmptyTransformer(Transformer):
    """
    A transformer that handles empty columns in a DataFrame.

    This transformer identifies and processes columns that contain no data (empty columns) in a given DataFrame.
    It can remove these columns during the conversion process and restore them during the reverse conversion process.

    Attributes:
        empty_columns (list): A list of column names that are identified as empty.

    Methods:
        fit(metadata: Metadata | None = None, **kwargs: dict[str, Any]):
            Fits the transformer to the data by identifying empty columns based on provided metadata.
        convert(raw_data: pd.DataFrame) -> pd.DataFrame:
            Converts the raw data by removing the identified empty columns.
        reverse_convert(processed_data: pd.DataFrame) -> pd.DataFrame:
            Reverses the conversion by restoring the previously removed empty columns.
    """

    empty_columns: set
    """
    Set of column names that are identified as empty. This attribute is populated during the fitting process
    and is used to remove these columns during the conversion process and restore them during the reverse conversion process.
    """

    def __init__(self):
        self.empty_columns = set()

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the transformer.
        Remember the empty_columns from all columns.

        Args:
            metadata (Metadata | None): The metadata containing information about the data, including empty columns.
            **kwargs (dict[str, Any]): Additional keyword arguments.

        Returns:
            None
        """
        for each_col in metadata.get("empty_columns"):
            if metadata.get_column_data_type(each_col) == "empty":
                self.empty_columns.add(each_col)

        logger.info("EmptyTransformer Fitted.")

        self.fitted = True

        return

    def convert(self, raw_data: pd.DataFrame) -> pd.DataFrame:
        """
        Converts the raw data by removing the identified empty columns.

        Args:
            raw_data (pd.DataFrame): The input DataFrame containing the raw data.

        Returns:
            pd.DataFrame: The processed DataFrame with empty columns removed.
        """
        processed_data = raw_data

        logger.info("Converting data using EmptyTransformer...")

        for each_col in self.empty_columns:
            processed_data = self.remove_columns(processed_data, [each_col])
        logger.info("Converting data using EmptyTransformer... Finished (No action).")

        return processed_data

    def reverse_convert(self, processed_data: pd.DataFrame) -> pd.DataFrame:
        """
        Reverses the conversion by restoring the previously removed empty columns.

        Args:
            processed_data (pd.DataFrame): The input DataFrame containing the processed data.

        Returns:
            pd.DataFrame: The DataFrame with previously removed empty columns restored.
        """

        if not self.fitted or not self.empty_columns:
            return processed_data

        for col_name in self.empty_columns:
            # Create an empty column with the same number of rows as `processed_data`.
            empty_df = pd.DataFrame({col_name: [None] * len(processed_data)})
            processed_data = self.attach_columns(processed_data, empty_df)

        return processed_data


@hookimpl
def register(manager):
    manager.register("EmptyTransformer", EmptyTransformer)



from __future__ import annotations

from sdgx.cachers.base import Cacher


class MyOwnCache(Cacher): ...


from sdgx.cachers.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("DummyCache", MyOwnCache)



"""StudentTUnivariate module."""

from scipy.stats import t

from sdgx.models.components.sdv_copulas.univariate.base import (
    BoundedType,
    ParametricType,
    ScipyModel,
)


class StudentTUnivariate(ScipyModel):
    """Wrapper around scipy.stats.t.

    Documentation: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.t.html
    """

    PARAMETRIC = ParametricType.PARAMETRIC
    BOUNDED = BoundedType.UNBOUNDED

    MODEL_CLASS = t

    def _fit_constant(self, X):
        self._fit(X)
        self._params["scale"] = 0

    def _fit(self, X):
        dataframe, loc, scale = t.fit(X)
        self._params = {"df": dataframe, "loc": loc, "scale": scale}

    def _is_constant(self):
        return self._params["scale"] == 0

    def _extract_constant(self):
        return self._params["loc"]



import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.filter.positive_negative import PositiveNegativeFilter


@pytest.fixture
def pos_neg_test_df():
    row_cnt = 1000
    header = ["int_id", "pos_int", "neg_int", "pos_float", "neg_float", "mixed_int", "mixed_float"]

    np.random.seed(42)
    int_id = list(range(row_cnt))
    pos_int = np.random.randint(1, 100, size=row_cnt)
    neg_int = np.random.randint(-100, 0, size=row_cnt)
    pos_float = np.random.uniform(0, 100, size=row_cnt)
    neg_float = np.random.uniform(-100, 0, size=row_cnt)
    mixed_int = np.random.randint(-50, 50, size=row_cnt)
    mixed_float = np.random.uniform(-50, 50, size=row_cnt)

    X = [
        [
            int_id[i],
            pos_int[i],
            neg_int[i],
            pos_float[i],
            neg_float[i],
            mixed_int[i],
            mixed_float[i],
        ]
        for i in range(row_cnt)
    ]

    yield pd.DataFrame(X, columns=header)


def test_positive_negative_filter(pos_neg_test_df: pd.DataFrame):
    # Get metadata
    metadata_df = Metadata.from_dataframe(pos_neg_test_df)

    # Initialize PositiveNegativeFilter
    pos_neg_filter = PositiveNegativeFilter()
    assert not pos_neg_filter.fitted

    # Test fit method
    pos_neg_filter.fit(metadata_df)
    assert pos_neg_filter.fitted
    assert pos_neg_filter.positive_columns == {"int_id", "pos_int", "pos_float"}
    assert pos_neg_filter.negative_columns == {"neg_int", "neg_float"}

    # Test convert method
    converted_df = pos_neg_filter.convert(pos_neg_test_df)
    assert converted_df.shape == pos_neg_test_df.shape
    assert (converted_df["pos_int"] >= 0).all()
    assert (converted_df["pos_float"] >= 0).all()
    assert (converted_df["neg_int"] <= 0).all()
    assert (converted_df["neg_float"] <= 0).all()

    # Test reverse_convert method
    reverse_converted_df = pos_neg_filter.reverse_convert(converted_df)
    assert reverse_converted_df.shape[1] == converted_df.shape[1]
    assert (reverse_converted_df["pos_int"] >= 0).all()
    assert (reverse_converted_df["pos_float"] >= 0).all()
    assert (reverse_converted_df["neg_int"] <= 0).all()
    assert (reverse_converted_df["neg_float"] <= 0).all()

    # Check: whether mixed columns remained unchanged
    pd.testing.assert_series_equal(pos_neg_test_df["mixed_int"], reverse_converted_df["mixed_int"])
    pd.testing.assert_series_equal(
        pos_neg_test_df["mixed_float"], reverse_converted_df["mixed_float"]
    )

    # Check if reverse_convert correctly filtered out non-compliant rows (samples)
    assert reverse_converted_df.shape[0] <= pos_neg_test_df.shape[0]


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import os
import shutil
from functools import cached_property
from pathlib import Path
from typing import Generator
from uuid import uuid4

import numpy as np
import pandas as pd
from numpy import ndarray

from sdgx.data_connectors.dataframe_connector import DataFrameConnector
from sdgx.data_loader import DataLoader

DEFAULT_CACHE_ROOT = os.getenv("SDG_NDARRAY_CACHE_ROOT", "./.ndarry_cache")


class NDArrayLoader:
    """
    Cache ndarray in disk, allow slice and random access.

    Support for storing two-dimensional data by columns.
    """

    def __init__(self, cache_root: str | Path = DEFAULT_CACHE_ROOT, save_to_file=True) -> None:
        self.store_index = 0
        self.cache_root = Path(cache_root).expanduser().resolve()
        self.save_to_file = save_to_file
        if save_to_file:
            self.cache_root.mkdir(exist_ok=True, parents=True)
        else:
            self.ndarray_list = []

    @staticmethod
    def get_auto_save(raw_data) -> NDArrayLoader:
        save_to_file = True
        if isinstance(raw_data, pd.DataFrame) or (
            isinstance(raw_data, DataLoader)
            and isinstance(raw_data.data_connector, DataFrameConnector)
        ):
            save_to_file = False
        return NDArrayLoader(save_to_file=save_to_file)

    @cached_property
    def subdir(self) -> str:
        """
        Prevent collision of cache files.
        """
        return uuid4().hex

    @cached_property
    def cache_dir(self) -> Path:
        """Cache directory for storing ndarray."""

        return self.cache_root / self.subdir

    def _get_cache_filename(self, index: int) -> Path:
        return self.cache_dir / f"{index}.npy"

    def store(self, ndarray: ndarray):
        """
        Spliting and storing columns of ndarry to disk, one by one.
        """
        if self.save_to_file:
            self.cache_dir.mkdir(exist_ok=True, parents=True)
            for ndarray in np.split(ndarray, indices_or_sections=ndarray.shape[1], axis=1):
                np.save(self._get_cache_filename(self.store_index), ndarray)
                self.store_index += 1
        else:
            for ndarray in np.split(ndarray, indices_or_sections=ndarray.shape[1], axis=1):
                self.ndarray_list.append(ndarray)
                self.store_index += 1

    def load(self, index: int) -> ndarray:
        """
        Load ndarray from disk by index of column.
        """
        if self.save_to_file:
            return np.load(self._get_cache_filename(int(index)))
        else:
            return self.ndarray_list[index]

    def cleanup(self):
        if self.save_to_file:
            try:
                shutil.rmtree(self.cache_dir, ignore_errors=True)
            except AttributeError:
                pass
        self.store_index = 0

    def iter(self) -> Generator[ndarray, None, None]:
        for i in range(self.store_index):
            yield self.load(i)

    def get_all(self) -> ndarray:
        return np.concatenate([array for array in self.iter()], axis=1)

    @cached_property
    def __shape_0(self):
        return self.load(0).shape[0]

    @property
    def shape(self) -> tuple[int, int]:
        return self.__shape_0, self.store_index

    def __len__(self):
        return self.shape[0]

    def __getitem__(self, key: int | slice | tuple[int | slice, int | slice]):
        if not isinstance(key, tuple):
            # NDArrayLoader[:], NDArrayLoader[1]
            return np.concatenate([self.load(i)[key] for i in range(self.store_index)], axis=1)
        else:
            # NDArrayLoader[:, 1], NDArrayLoader[1, :], NDArrayLoader[:, :]
            x_slice, y_slice = key
            if not isinstance(y_slice, slice):
                # NDArrayLoader[:, 1]
                return self.load(y_slice)[x_slice].squeeze(axis=1)
            if not isinstance(x_slice, slice):
                # NDArrayLoader[1, :]
                return np.concatenate(
                    [
                        self.load(i)[x_slice]
                        for i in range(
                            y_slice.start or 0,
                            min(y_slice.stop or self.store_index, self.store_index),
                            y_slice.step or 1,
                        )
                    ],
                )
            else:
                # NDArrayLoader[:, :]
                return np.concatenate(
                    [
                        self.load(i)[x_slice]
                        for i in range(
                            y_slice.start or 0,
                            min(y_slice.stop or self.store_index, self.store_index),
                            y_slice.step or 1,
                        )
                    ],
                    axis=1,
                )

    def __del__(self):
        self.cleanup()



import random

import numpy as np
import pandas as pd
import pytest

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.transformers.column_order import ColumnOrderTransformer
from sdgx.data_processors.transformers.discrete import DiscreteTransformer


@pytest.fixture
def df_data():
    row_cnt = 1000
    header = ["int_id", "discrete_val", "int_random", "bool_random", "float_random"]

    int_id = list(range(row_cnt))
    discrete_val = list(random.choice(["a", "b", "c"]) for _ in range(row_cnt))

    int_random = np.random.randint(100, size=row_cnt)
    bool_random = int_random < 50
    float_random = np.random.randn(row_cnt)

    X = [
        [int_id[i], discrete_val[i], int_random[i], bool_random[i], float_random[i]]
        for i in range(row_cnt)
    ]
    # Convert the list of lists to a DataFrame
    df = pd.DataFrame(X, columns=header)
    yield df


def is_a_string_list(lst):
    """
    Check if all items in a list are strings.

    Parameters:
    lst (list): The list to check.

    Returns:
    bool: True if all items in the list are strings, False otherwise.
    """
    return all(isinstance(item, str) for item in lst)


def is_an_integer_list(lst):
    """
    Check if all elements in the list are integers or floats that are also integers.

    Parameters:
    lst (list): The list to be checked.

    Returns:
    bool: True if all elements are integers or floats that are also integers, False otherwise.
    """
    return all(isinstance(i, int) or (isinstance(i, float) and i.is_integer()) for i in lst)


def test_discrete_transformer_fit_test_df(df_data: pd.DataFrame):
    """
    Test the fit and convert methods of the DiscreteTransformer class.

    This function tests the following:
    1. The fit method of the DiscreteTransformer class.
    2. The convert method of the DiscreteTransformer class.
    3. The reverse_convert method of the DiscreteTransformer class.
    4. The equality of the original dataframe and the reversely converted dataframe.

    Parameters:
    df_data (pd.DataFrame): The input dataframe to be tested.

    Returns:
    None
    """
    # get metadata
    metadata_df = Metadata.from_dataframe(df_data)

    # use another ColumnOrderTransformer to reorder the columns
    order_transformer = ColumnOrderTransformer()
    order_transformer.fit(metadata_df)

    # fit the transformer
    transformer = DiscreteTransformer()
    assert not transformer.fitted
    transformer.fit(metadata_df, df_data)
    assert transformer.fitted
    assert transformer.discrete_columns == {"discrete_val"}

    # convert df using the fitted transformer
    converted_df = transformer.convert(df_data)
    assert isinstance(converted_df, pd.DataFrame)
    assert is_an_integer_list(converted_df["discrete_val_a"].to_list())
    assert is_an_integer_list(converted_df["discrete_val_b"].to_list())
    assert is_an_integer_list(converted_df["discrete_val_c"].to_list())

    # reverse convert the df back to original
    reverse_converted_df = transformer.reverse_convert(converted_df)
    reverse_converted_df = order_transformer.reverse_convert(reverse_converted_df)
    assert isinstance(reverse_converted_df, pd.DataFrame)
    assert is_a_string_list(reverse_converted_df["discrete_val"].to_list())

    # check if the dataframe is equal to the original one
    # use the eq method and .all().all() to check the equality of two 2d dataframes
    assert reverse_converted_df.eq(df_data).all().all()



from __future__ import annotations

from sdgx.data_connectors.base import DataConnector


class MyOwnDataConnector(DataConnector): ...


from sdgx.data_connectors.extension import hookimpl


@hookimpl
def register(manager):
    manager.register("DummyDataConnector", MyOwnDataConnector)



import pandas as pd
import pytest

from sdgx.data_models.inspectors.fixed_combination import FixedCombinationInspector


@pytest.fixture
def test_fixed_combination_data():
    data = {
        "A": [1, 2, 3, 4, 5],
        "B": [2, 4, 6, 8, 10],  # B is 2 * A
        "C": [5, 5, 5, 5, 5],  # C is constant
        "D": [1, 3, 5, 7, 9],  # D is not related to A or B
        "E": [2, 4, 6, 8, 10],  # B is 2 * A
        "categorical_1": ["apple", "banana", "apple", "banana", "cherry"],  # fruits
        "categorical_2": ["red", "yellow", "red", "yellow", "pink"],  # colors
        "categorical_3": [1, 2, 1, 2, 3],  # int to string
        "categorical_4": ["one", "two", "one", "two", "three"],  # int to string
        "categorical_5": [0.1, 0.5, 0.1, 0.5, 1.0],  # float to string
        "categorical_6": ["light", "medium", "light", "medium", "heavy"],  # weight descriptors
    }
    df = pd.DataFrame(data)
    yield df


def test_fixed_combination_inspector(test_fixed_combination_data: pd.DataFrame):
    inspector = FixedCombinationInspector()
    inspector.fit(test_fixed_combination_data)
    assert inspector.ready
    assert inspector.fixed_combinations

    expected_combinations = {
        "A": {"categorical_3", "D", "E", "B"},
        "B": {"categorical_3", "D", "E", "A"},
        "D": {"categorical_3", "E", "A", "B"},
        "E": {"categorical_3", "D", "A", "B"},
        "categorical_3": {"categorical_4", "D", "E", "A", "B"},
        "categorical_1": {"categorical_2"},
        "categorical_5": {"categorical_6"},
    }

    assert inspector.fixed_combinations == expected_combinations
    assert inspector.inspect_level == 70


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



import pandas as pd
import pytest

from sdgx.models.statistics.single_table.copula import GaussianCopulaSynthesizerModel


@pytest.fixture
def dummy_data(dummy_single_table_path):
    yield pd.read_csv(dummy_single_table_path)


def test_gaussian_copula(dummy_single_table_metadata, dummy_single_table_data_loader):
    model = GaussianCopulaSynthesizerModel()
    model.fit(dummy_single_table_metadata, dummy_single_table_data_loader)

    sampled_data = model.sample(10)
    original_data = dummy_single_table_data_loader.load_all()
    assert len(sampled_data) == 10
    assert sampled_data.columns.tolist() == original_data.columns.tolist()



from __future__ import annotations

from typing import Any

from pandas import DataFrame

from sdgx.data_models.metadata import Metadata
from sdgx.data_processors.extension import hookimpl
from sdgx.data_processors.transformers.base import Transformer
from sdgx.utils import logger


class OutlierTransformer(Transformer):
    """
    A transformer class to handle outliers in the data by converting them to specified fill values.

    Attributes:
        int_columns (set): A set of column names that contain integer values.
        int_outlier_fill_value (int): The value to fill in for outliers in integer columns. Default is 0.
        float_columns (set): A set of column names that contain float values.
        float_outlier_fill_value (float): The value to fill in for outliers in float columns. Default is 0.
    """

    int_columns: set
    """
    set: A set of column names that contain integer values. These columns will have their outliers replaced by `int_outlier_fill_value`.
    """

    int_outlier_fill_value: int
    """
    int: The value to fill in for outliers in integer columns. Default is 0.
    """

    float_columns: set
    """
    set: A set of column names that contain float values. These columns will have their outliers replaced by `float_outlier_fill_value`.
    """

    float_outlier_fill_value: float
    """
    float: The value to fill in for outliers in float columns. Default is 0.
    """

    def __init__(self):
        self.int_columns = set()
        self.int_outlier_fill_value = 0
        self.float_columns = set()
        self.float_outlier_fill_value = float(0)

    def fit(self, metadata: Metadata | None = None, **kwargs: dict[str, Any]):
        """
        Fit method for the transformer.

        Records the names of integer and float columns from the metadata.

        Args:
            metadata (Metadata | None): The metadata object containing column type information.
            **kwargs: Additional keyword arguments.
        """
        # int columns
        for each_col in metadata.int_columns:
            if each_col not in metadata.column_list:
                continue
            if metadata.get_column_data_type(each_col) == "int":
                self.int_columns.add(each_col)

        # float columns
        for each_col in metadata.float_columns:
            if each_col not in metadata.column_list:
                continue
            if metadata.get_column_data_type(each_col) == "float":
                self.float_columns.add(each_col)

        self.fitted = True

        logger.info("OutlierTransformer Fitted.")

    def convert(self, raw_data: DataFrame) -> DataFrame:
        """
        Convert method to handle outliers in the input data by replacing them with specified fill values.

        Args:
            raw_data (DataFrame): The input DataFrame containing the data to be processed.

        Returns:
            DataFrame: The processed DataFrame with outliers replaced by fill values.
        """
        res = raw_data

        logger.info("Converting data using OutlierTransformer...")

        # Dealing with the integer value columns
        def convert_to_int(value):
            try:
                return int(value)
            except ValueError:
                return self.int_outlier_fill_value

        for each_col in self.int_columns:
            res[each_col] = res[each_col].apply(convert_to_int)

        # Dealing with the float value columns
        def convert_to_float(value):
            try:
                return float(value)
            except ValueError:
                return self.float_outlier_fill_value

        for each_col in self.float_columns:
            res[each_col] = res[each_col].apply(convert_to_float)

        logger.info("Converting data using OutlierTransformer... Finished.")

        return res

    def reverse_convert(self, processed_data: DataFrame) -> DataFrame:
        """
        Reverse_convert method for the transformer (No action for OutlierTransformer).

        Args:
            processed_data (DataFrame): The processed DataFrame.

        Returns:
            DataFrame: The same processed DataFrame.
        """
        logger.info("Data reverse-converted by OutlierTransformer (No Action).")

        return processed_data


@hookimpl
def register(manager):
    """
    Register the OutlierTransformer with the manager.

    Args:
        manager: The manager object responsible for registering transformers.
    """
    manager.register("OutlierTransformer", OutlierTransformer)



# FixedCombinationInspector
from __future__ import annotations

from typing import Any

import pandas as pd

from sdgx.data_models.inspectors.base import Inspector
from sdgx.data_models.inspectors.extension import hookimpl


class FixedCombinationInspector(Inspector):
    """
    FixedCombinationInspector is designed to identify columns in a DataFrame that have fixed relationships based on covariance.

    Attributes:
        fixed_combinations (dict[str, set[str]]): A dictionary mapping column names to sets of column names that have fixed relationships with them.
        _inspect_level (int): The inspection level for this inspector, set to 70.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fixed_combinations: dict[str, set[str]] = {}
        """
        A dictionary mapping column names to sets of column names that have fixed relationships with them.
        """

        self._inspect_level: int = 70
        """
        The inspection level for this inspector, set to 70. This attribute indicates the priority or depth of inspection that this inspector performs relative to other inspectors.
        """

    def fit(self, raw_data: pd.DataFrame, *args, **kwargs):
        """
        Fit the inspector to the raw data.
            Process fixed combinations of numerical and string columns:
            Numerical Columns: Calculate correlation using the covariance matrix.
            String Columns: Determine relationships based on one-to-one value mapping.
        """
        self.fixed_combinations = {}
        self._fit_numeric_relationships(raw_data)
        self._fit_one_to_one_relationships(raw_data)
        self.ready = True

    def _fit_numeric_relationships(self, raw_data: pd.DataFrame) -> None:
        """
        Calculate correlation using the covariance matrix.
        """
        # 1. Handle numeric relationships.
        numeric_columns = raw_data.select_dtypes(include=["int64", "float64"]).columns
        if len(numeric_columns) > 0:
            covariance_matrix = raw_data[numeric_columns].dropna().cov()
            for column in covariance_matrix.columns:
                related_columns = set(
                    covariance_matrix.index[covariance_matrix[column].abs() > 0.9]
                )
                related_columns.discard(column)
                if related_columns:
                    self.fixed_combinations[column] = related_columns

    def _fit_one_to_one_relationships(self, raw_data: pd.DataFrame) -> None:
        """
        Determine relationships based on one-to-one value mapping.
        """
        string_columns = raw_data.columns
        if len(string_columns) > 0:
            # Pre-compute the number of unique values for each column. - (Col_Name, Count) ...
            # Filter out potential PII columns and empty columns
            # Prioritizing those with a smaller number of unique values for processing.
            matched_columns = set()
            unique_counts = raw_data[string_columns].nunique(dropna=True)
            filter_condition = (unique_counts < (raw_data.shape[0] * 0.9)) & (unique_counts != 0)
            unique_counts = unique_counts[filter_condition]
            sorted_columns = unique_counts.sort_values().index.tolist()

            # For each unmatched column, check if there is a one-to-one correspondence.
            for i, col1 in enumerate(sorted_columns):
                if col1 in matched_columns:
                    continue
                for col2 in sorted_columns[i + 1 :]:
                    if col2 in matched_columns:
                        continue
                    if unique_counts[col1] != unique_counts[col2]:
                        continue
                    # For the two columns of data:
                    # 1. Remove the duplicate rows based on their combination
                    # 2. And then check if there are any duplicate values in the remaining rows of both columns.
                    pairs = raw_data[[col1, col2]].dropna()
                    mapping_from_col = pairs.drop_duplicates(subset=[col1, col2])
                    duplicates_in_col1 = mapping_from_col.duplicated(subset=col1, keep=False)
                    duplicates_in_col2 = mapping_from_col.duplicated(subset=col2, keep=False)

                    # Save the relationships when both columns have no duplicate values.
                    if not duplicates_in_col1.any() and not duplicates_in_col2.any():
                        if col1 not in self.fixed_combinations:
                            self.fixed_combinations[col1] = set()
                        self.fixed_combinations[col1].add(col2)
                        matched_columns.add(col1)
                        matched_columns.add(col2)
                        break

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""

        return {"fixed_combinations": self.fixed_combinations}


@hookimpl
def register(manager):
    manager.register("FixCombinationInspector", FixedCombinationInspector)



import copy
import json
import re
import sys
from functools import wraps

import click
import importlib_metadata

from sdgx.cli.message import ExceptionMessage, NormalMessage
from sdgx.log import LOG_TO_FILE, add_log_file_handler, logger
from sdgx.utils import find_free_port


def cli_wrapper(func):
    @click.option("--json_output", type=bool, default=False, help="Exit with json output.")
    @click.option("--log_to_file", type=bool, default=False, help="Log to file.")
    @wraps(func)
    def wrapper(json_output, log_to_file, *args, **kwargs):
        if log_to_file and not LOG_TO_FILE:
            add_log_file_handler()
        try:
            func(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
            if json_output:
                ExceptionMessage.from_exception(e).send()
            exit(getattr(e, "EXIT_CODE", -1))
        else:
            if json_output:
                NormalMessage().send()
            exit(0)

    return wrapper


def load_entry_point(distribution, group, name):
    dist_obj = importlib_metadata.distribution(distribution)
    eps = [ep for ep in dist_obj.entry_points if ep.group == group and ep.name == name]
    if not eps:
        raise ImportError("Entry point %r not found" % ((group, name),))
    return eps[0].load()


def torch_run_warpper(func):
    """
    Experimental feature for native torchrun.

    Alternatively, people can use `torchrun $(which sdgx)`

    FIXME: This is not compatible with click.testing.CliRunner
    """

    @click.option("--torchrun", type=bool, default=False, help="Use `torchrun` to run cli.")
    @click.option(
        "--torchrun_kwargs",
        type=str,
        default="{}",
        help="[Json String] torchrun kwargs.",
    )
    @wraps(func)
    def wrapper(torchrun, torchrun_kwargs, *args, **kwargs):
        if not torchrun:
            func(*args, **kwargs)
        else:
            torchrun_kwargs = json.loads(torchrun_kwargs)
            torchrun_kwargs.setdefault("master_port", find_free_port())
            origin_args = copy.deepcopy(sys.argv)
            sys.argv = [re.sub(r"(-script\.pyw|\.exe)?$", "", sys.argv[0])]
            for k, v in torchrun_kwargs.items():
                sys.argv.extend([f"--{k}", str(v)])
            # Remove [--torchrun=true] and [--torchrun, true] from origin_args
            if "--torchrun" in origin_args:
                i = origin_args.index("--torchrun")
                if i + 1 < len(origin_args) and origin_args[i + 1] == "true":
                    origin_args.pop(i)
                    origin_args.pop(i)

            if "--torchrun=true" in origin_args:
                origin_args.remove("--torchrun=true")

            sys.argv.extend(origin_args)
            sys.exit(load_entry_point("torch", "console_scripts", "torchrun")())

    return wrapper



"""Visualization utilities for the Copulas library."""

import pandas as pd

try:
    import matplotlib.pyplot as plt
except RuntimeError as e:
    if "Python is not installed as a framework." in e.message:
        import matplotlib

        matplotlib.use("PS")  # Avoid crash on macos
        import matplotlib.pyplot as plt


def scatter_3d(data, columns=None, fig=None, title=None, position=None):
    """Plot 3 dimensional data in a scatter plot."""
    fig = fig or plt.figure()
    position = position or 111

    ax = fig.add_subplot(position, projection="3d")
    ax.scatter(*(data[column] for column in columns or data.columns))
    if title:
        ax.set_title(title)
        ax.title.set_position([0.5, 1.05])

    return ax


def scatter_2d(data, columns=None, fig=None, title=None, position=None):
    """Plot 2 dimensional data in a scatter plot."""
    fig = fig or plt.figure()
    position = position or 111

    ax = fig.add_subplot(position)
    columns = columns or data.columns
    if len(columns) != 2:
        raise ValueError("Only 2 columns can be plotted")

    x, y = columns

    ax.scatter(data[x], data[y])
    plt.xlabel(x)
    plt.ylabel(y)

    if title:
        ax.set_title(title)
        ax.title.set_position([0.5, 1.05])

    return ax


def hist_1d(data, fig=None, title=None, position=None, bins=20, label=None):
    """Plot 1 dimensional data in a histogram."""
    fig = fig or plt.figure()
    position = position or 111

    ax = fig.add_subplot(position)
    ax.hist(data, density=True, bins=bins, alpha=0.8, label=label)

    if label:
        ax.legend()

    if title:
        ax.set_title(title)
        ax.title.set_position([0.5, 1.05])

    return ax


def side_by_side(plotting_func, arrays):
    """Plot side-by-side figures.

    Args:
        plotting_func (callable):
            A matplotlib function which takes in the standard plot kwargs.
        arrays (dict[str, np.ndarray]):
            A mapping from the name of the subplot to the values.
    """
    fig = plt.figure(figsize=(10, 4))

    position_base = f"1{len(arrays)}"
    for index, (title, array) in enumerate(arrays.items()):
        position = int(position_base + str(index + 1))
        plotting_func(array, fig=fig, title=title, position=position)

    plt.tight_layout()


def compare_3d(real, synth, columns=None, figsize=(10, 4)):
    """Generate a 3d scatter plot comparing real/synthetic data.

    Args:
        real (pd.DataFrame):
            The real data.
        synth (pd.DataFrame):
            The synthetic data.
        columns (list):
            The name of the columns to plot.
        figsize:
            Figure size, passed to matplotlib.
    """
    columns = columns or real.columns
    fig = plt.figure(figsize=figsize)

    scatter_3d(real[columns], fig=fig, title="Real Data", position=121)
    scatter_3d(synth[columns], fig=fig, title="Synthetic Data", position=122)

    plt.tight_layout()


def compare_2d(real, synth, columns=None, figsize=None):
    """Generate a 2d scatter plot comparing real/synthetic data.

    Args:
        real (pd.DataFrame):
            The real data.
        synth (pd.DataFrame):
            The synthetic data.
        columns (list):
            The name of the columns to plot.
        figsize:
            Figure size, passed to matplotlib.
    """
    x, y = columns or real.columns
    ax = real.plot.scatter(x, y, color="blue", alpha=0.5, figsize=figsize)
    ax = synth.plot.scatter(x, y, ax=ax, color="orange", alpha=0.5, figsize=figsize)
    ax.legend(["Real", "Synthetic"])


def compare_1d(real, synth, columns=None, figsize=None):
    """Generate a 1d scatter plot comparing real/synthetic data.

    Args:
        real (pd.DataFrame):
            The real data.
        synth (pd.DataFrame):
            The synthetic data.
        columns (list):
            The name of the columns to plot.
        figsize:
            Figure size, passed to matplotlib.
    """
    if len(real.shape) == 1:
        real = pd.DataFrame({"": real})
        synth = pd.DataFrame({"": synth})

    columns = columns or real.columns

    num_cols = len(columns)
    fig_cols = min(2, num_cols)
    fig_rows = (num_cols // fig_cols) + 1
    prefix = f"{fig_rows}{fig_cols}"

    figsize = figsize or (5 * fig_cols, 3 * fig_rows)
    fig = plt.figure(figsize=figsize)

    for idx, column in enumerate(columns):
        position = int(prefix + str(idx + 1))
        hist_1d(real[column], fig=fig, position=position, title=column, label="Real")
        hist_1d(synth[column], fig=fig, position=position, title=column, label="Synthetic")

    plt.tight_layout()



"""
    GaussianMultivariate module.
    需要对 copulas 的代码进行修改 + 性能优化
"""

import logging
import sys
import warnings

import numpy as np
import pandas as pd
from scipy import stats

from sdgx.models.components.sdv_copulas import (
    EPSILON,
    check_valid_values,
    get_instance,
    get_qualified_name,
    random_state,
    store_args,
    validate_random_state,
)
from sdgx.models.components.sdv_copulas.multivariate.base import Multivariate
from sdgx.models.components.sdv_copulas.univariate import GaussianUnivariate, Univariate

LOGGER = logging.getLogger(__name__)
DEFAULT_DISTRIBUTION = Univariate


class GaussianMultivariate(Multivariate):
    """Class for a multivariate distribution that uses the Gaussian copula.

    Args:
        distribution (str or dict):
            Fully qualified name of the class to be used for modeling the marginal
            distributions or a dictionary mapping column names to the fully qualified
            distribution names.
    """

    correlation = None
    columns = None
    univariates = None

    @store_args
    def __init__(self, distribution=DEFAULT_DISTRIBUTION, random_state=None):
        self.random_state = validate_random_state(random_state)
        self.distribution = distribution

    def __repr__(self):
        """Produce printable representation of the object."""
        if self.distribution == DEFAULT_DISTRIBUTION:
            distribution = ""
        elif isinstance(self.distribution, type):
            distribution = f'distribution="{self.distribution.__name__}"'
        else:
            distribution = f'distribution="{self.distribution}"'

        return f"GaussianMultivariate({distribution})"

    def _transform_to_normal(self, X):
        if isinstance(X, pd.Series):
            X = X.to_frame().T
        elif not isinstance(X, pd.DataFrame):
            if len(X.shape) == 1:
                X = [X]

            X = pd.DataFrame(X, columns=self.columns)

        U = []
        for column_name, univariate in zip(self.columns, self.univariates):
            if column_name in X:
                column = X[column_name]
                U.append(univariate.cdf(column.to_numpy()).clip(EPSILON, 1 - EPSILON))

        return stats.norm.ppf(np.column_stack(U))

    def _get_correlation(self, X):
        """Compute correlation matrix with transformed data.

        Args:
            X (numpy.ndarray):
                Data for which the correlation needs to be computed.

        Returns:
            numpy.ndarray:
                computed correlation matrix.
        """
        result = self._transform_to_normal(X)
        correlation = pd.DataFrame(data=result).corr().to_numpy()
        correlation = np.nan_to_num(correlation, nan=0.0)
        # If singular, add some noise to the diagonal
        if np.linalg.cond(correlation) > 1.0 / sys.float_info.epsilon:
            correlation = correlation + np.identity(correlation.shape[0]) * EPSILON

        return pd.DataFrame(correlation, index=self.columns, columns=self.columns)

    @check_valid_values
    def fit(self, X):
        """Compute the distribution for each variable and then its correlation matrix.

        Arguments:
            X (pandas.DataFrame):
                Values of the random variables.
        """
        LOGGER.info("Fitting %s", self)

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)

        columns = []
        univariates = []
        for column_name, column in X.items():
            if isinstance(self.distribution, dict):
                distribution = self.distribution.get(column_name, DEFAULT_DISTRIBUTION)
            else:
                distribution = self.distribution

            LOGGER.debug("Fitting column %s to %s", column_name, distribution)

            univariate = get_instance(distribution)
            try:
                univariate.fit(column)
            except BaseException:
                warning_message = (
                    f"Unable to fit to a {distribution} distribution for column {column_name}. "
                    "Using a Gaussian distribution instead."
                )
                warnings.warn(warning_message)
                univariate = GaussianUnivariate()
                univariate.fit(column)

            columns.append(column_name)
            univariates.append(univariate)

        self.columns = columns
        self.univariates = univariates

        LOGGER.debug("Computing correlation")
        self.correlation = self._get_correlation(X)
        self.fitted = True

        LOGGER.debug("GaussianMultivariate fitted successfully")

    def probability_density(self, X):
        """Compute the probability density for each point in X.

        Arguments:
            X (pandas.DataFrame):
                Values for which the probability density will be computed.

        Returns:
            numpy.ndarray:
                Probability density values for points in X.

        Raises:
            NotFittedError:
                if the model is not fitted.
        """
        self.check_fit()
        transformed = self._transform_to_normal(X)
        return stats.multivariate_normal.pdf(transformed, cov=self.correlation)

    def cumulative_distribution(self, X):
        """Compute the cumulative distribution value for each point in X.

        Arguments:
            X (pandas.DataFrame):
                Values for which the cumulative distribution will be computed.

        Returns:
            numpy.ndarray:
                Cumulative distribution values for points in X.

        Raises:
            NotFittedError:
                if the model is not fitted.
        """
        self.check_fit()
        transformed = self._transform_to_normal(X)
        return stats.multivariate_normal.cdf(transformed, cov=self.correlation)

    def _get_conditional_distribution(self, conditions):
        """Compute the parameters of a conditional multivariate normal distribution.

        The parameters of the conditioned distribution are computed as specified here:
        https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Conditional_distributions

        Args:
            conditions (pandas.Series):
                Mapping of the column names and column values to condition on.
                The input values have already been transformed to their normal distribution.

        Returns:
            tuple:
                * means (numpy.array):
                    mean values to use for the conditioned multivariate normal.
                * covariance (numpy.array):
                    covariance matrix to use for the conditioned
                  multivariate normal.
                * columns (list):
                    names of the columns that will be sampled conditionally.
        """
        columns2 = conditions.index
        columns1 = self.correlation.columns.difference(columns2)

        sigma11 = self.correlation.loc[columns1, columns1].to_numpy()
        sigma12 = self.correlation.loc[columns1, columns2].to_numpy()
        sigma21 = self.correlation.loc[columns2, columns1].to_numpy()
        sigma22 = self.correlation.loc[columns2, columns2].to_numpy()

        mu1 = np.zeros(len(columns1))
        mu2 = np.zeros(len(columns2))

        sigma12sigma22inv = sigma12 @ np.linalg.inv(sigma22)

        mu_bar = mu1 + sigma12sigma22inv @ (conditions - mu2)
        sigma_bar = sigma11 - sigma12sigma22inv @ sigma21

        return mu_bar, sigma_bar, columns1

    def _get_normal_samples(self, num_rows, conditions):
        """Get random rows in the standard normal space.

        If no conditions are given, the values are sampled from a standard normal
        multivariate.

        If conditions are given, they are transformed to their equivalent standard
        normal values using their marginals and then the values are sampled from
        a standard normal multivariate conditioned on the given condition values.
        """
        if conditions is None:
            covariance = self.correlation
            columns = self.columns
            means = np.zeros(len(columns))
        else:
            conditions = pd.Series(conditions)
            normal_conditions = self._transform_to_normal(conditions)[0]
            normal_conditions = pd.Series(normal_conditions, index=conditions.index)
            means, covariance, columns = self._get_conditional_distribution(normal_conditions)

        samples = np.random.multivariate_normal(means, covariance, size=num_rows)
        return pd.DataFrame(samples, columns=columns)

    @random_state
    def sample(self, num_rows=1, conditions=None):
        """Sample values from this model.

        Argument:
            num_rows (int):
                Number of rows to sample.
            conditions (dict or pd.Series):
                Mapping of the column names and column values to condition on.

        Returns:
            numpy.ndarray:
                Array of shape (n_samples, *) with values randomly
                sampled from this model distribution. If conditions have been
                given, the output array also contains the corresponding columns
                populated with the given values.

        Raises:
            NotFittedError:
                if the model is not fitted.
        """
        self.check_fit()

        samples = self._get_normal_samples(num_rows, conditions)

        output = {}
        for column_name, univariate in zip(self.columns, self.univariates):
            if conditions and column_name in conditions:
                # Use the values that were given as conditions in the original space.
                output[column_name] = np.full(num_rows, conditions[column_name])
            else:
                cdf = stats.norm.cdf(samples[column_name])
                output[column_name] = univariate.percent_point(cdf)

        return pd.DataFrame(data=output)

    def to_dict(self):
        """Return a `dict` with the parameters to replicate this object.

        Returns:
            dict:
                Parameters of this distribution.
        """
        self.check_fit()
        univariates = [univariate.to_dict() for univariate in self.univariates]

        return {
            "correlation": self.correlation.to_numpy().tolist(),
            "univariates": univariates,
            "columns": self.columns,
            "type": get_qualified_name(self),
        }

    @classmethod
    def from_dict(cls, copula_dict):
        instance = cls()
        instance.univariates = []
        columns = copula_dict["columns"]
        instance.columns = columns

        for parameters in copula_dict["univariates"]:
            instance.univariates.append(Univariate.from_dict(parameters))

        correlation = copula_dict["correlation"]
        instance.correlation = pd.DataFrame(correlation, index=columns, columns=columns)
        instance.fitted = True

        return instance



import numpy as np
import pytest

from sdgx.models.components.optimize.ndarray_loader import NDArrayLoader


@pytest.fixture
def ndarray_loaders(tmp_path, ndarray_list):
    def generator():
        cache_dir = tmp_path / "ndarrycache"
        for save in [True, False]:
            loader = NDArrayLoader(cache_root=cache_dir, save_to_file=save)
            for ndarray in ndarray_list:
                loader.store(ndarray)
            yield loader
            loader.cleanup()

    yield generator()


@pytest.fixture
def ndarray_list():
    """
    1, 4, 7
    2, 5, 8
    3, 6, 9
    """
    yield [
        np.array([[1], [2], [3]]),
        np.array([[4], [5], [6]]),
        np.array([[7], [8], [9]]),
    ]


def subtest_ndarray_loader_function(ndarray_loader: NDArrayLoader, ndarray_list):
    ndarray_all = np.concatenate(ndarray_list, axis=1)

    for i, ndarray in enumerate(ndarray_loader.iter()):
        np.testing.assert_equal(ndarray, ndarray_list[i])
    np.testing.assert_equal(ndarray_loader.get_all(), ndarray_all)

    assert ndarray_loader.shape == ndarray_all.shape


def subtest_ndarray_loader_slice(ndarray_loader: NDArrayLoader, ndarray_list):
    ndarray_all = np.concatenate(ndarray_list, axis=1)

    np.testing.assert_equal(ndarray_loader[:], ndarray_all[:])
    np.testing.assert_equal(ndarray_loader[::], ndarray_all[::])
    np.testing.assert_equal(ndarray_loader[:, :], ndarray_all[:, :])
    np.testing.assert_equal(ndarray_loader[::, ::], ndarray_all[::, ::])
    np.testing.assert_equal(ndarray_loader[:, 1], ndarray_all[:, 1])
    np.testing.assert_equal(ndarray_loader[1, :], ndarray_all[1, :])

    np.testing.assert_equal(ndarray_loader[1:3], ndarray_all[1:3])
    """
    2, 3
    5, 6
    8, 9
    """

    np.testing.assert_equal(ndarray_loader[1:3, 1], ndarray_all[1:3, 1])
    """
    5
    6
    """

    np.testing.assert_equal(ndarray_loader[1:3, 1:3], ndarray_all[1:3, 1:3])
    """
    5, 6
    8, 9
    """


def test_ndarray_loader(ndarray_list, ndarray_loaders):
    for ndarray_loader in ndarray_loaders:
        subtest_ndarray_loader_function(ndarray_loader, ndarray_list)
        subtest_ndarray_loader_slice(ndarray_loader, ndarray_list)


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

import glob
import importlib
from os.path import basename, dirname, isfile, join
from typing import Any

import pluggy

from sdgx import models
from sdgx.exceptions import InitializationError, NotFoundError, RegisterError
from sdgx.utils import Singleton, logger


class Manager(metaclass=Singleton):
    """
    Base class for all manager.

    Manager is a singleton class for preventing multiple initialization.

    Define following attributes in subclass:
        * register_type: Base class for registered class
        * project_name: Name of entry-point for extensio
        * hookspecs_model: Hook specification model(where @hookspec is defined)

    For available managers, please refer to :ref:`Plugin-supported modules`

    """

    register_type: type = object
    """
    Base class for registered class
    """
    project_name: str = ""
    """
    Name of entry-point for extension
    """

    hookspecs_model = None
    """
    Hook specification model(where @hookspec is defined)
    """

    def __init__(self):
        self.pm = pluggy.PluginManager(self.project_name)
        self.pm.add_hookspecs(self.hookspecs_model)
        self._registed_cls: dict[str, type[self.register_type]] = {}

        # Load all
        self.pm.load_setuptools_entrypoints(self.project_name)

        # Load all local model
        self.load_all_local_model()

    def load_all_local_model(self):
        """
        Implement this function to load all local model
        """
        return

    @property
    def registed_cls(self) -> dict[str, type]:
        """
        Access all registed class.

        Lazy load, only load once.
        """
        if self._registed_cls:
            return self._registed_cls
        for f in self.pm.hook.register(manager=self):
            try:
                f()
            except Exception as e:
                logger.exception(RegisterError(e))
                continue
        return self._registed_cls

    def _load_dir(self, module):
        """
        Import all python files in a submodule.
        """
        modules = glob.glob(join(dirname(module.__file__), "*.py"))
        sub_packages = (
            basename(f)[:-3] for f in modules if isfile(f) and not f.endswith("__init__.py")
        )
        packages = (str(module.__package__) + "." + i for i in sub_packages)
        for p in packages:
            self.pm.register(importlib.import_module(p))

    def _normalize_name(self, name: str) -> str:
        return name.strip().lower()

    def register(self, cls_name, cls: type):
        """
        Register a new model, if the model is already registed, skip it.
        """

        cls_name = self._normalize_name(cls_name)
        logger.debug(f"Register for new model: {cls_name}")
        if cls in self._registed_cls.values():
            logger.error(f"SKIP: {cls_name} is already registed")
            return
        if not issubclass(cls, self.register_type):
            logger.error(f"SKIP: {cls_name} is not a subclass of {self.register_type}")
            return
        self._registed_cls[cls_name] = cls

    def init(self, c, **kwargs: dict[str, Any]):
        """
        Init a new subclass of self.register_type.

        Raises:
            NotFoundError: if cls_name is not registered
            InitializationError: if failed to initialize
        """
        if isinstance(c, self.register_type):
            return c

        if isinstance(c, type):
            cls_type = c
        else:
            c = self._normalize_name(c)

            if not c in self.registed_cls:
                raise NotFoundError
            cls_type = self.registed_cls[c]
        try:
            instance = cls_type(**kwargs)
            if not isinstance(instance, self.register_type):
                raise InitializationError(f"{c} is not a subclass of {self.register_type}.")
            return instance
        except Exception as e:
            raise InitializationError(e)



from __future__ import annotations

import pytest

from sdgx.data_connectors.csv_connector import CsvConnector


@pytest.fixture
def csv_file(tmp_path):
    csv = tmp_path / "test.csv"
    csv.write_text("index,a,b,c\n0,1,2,3\n1,4,5,6")
    yield csv
    csv.unlink()


@pytest.fixture
def csv_connector(csv_file):
    return CsvConnector(
        path=csv_file,
    )


def test_read(csv_connector: CsvConnector):
    df = csv_connector.read()
    assert len(df) == 2

    df = csv_connector.read(offset=1)
    assert len(df) == 1

    df = csv_connector.read(limit=1)
    assert len(df) == 1

    df = csv_connector.read(offset=1, limit=1)
    assert len(df) == 1

    df = csv_connector.read(offset=1, limit=2)
    assert len(df) == 1

    df = csv_connector.read(offset=100)
    assert len(df) == 0


def test_columns(csv_connector: CsvConnector):
    columns = csv_connector.columns()
    assert isinstance(columns, list)
    assert len(columns) == 4


def test_keys(csv_connector: CsvConnector):
    keys = csv_connector.keys()
    assert isinstance(keys, list)
    assert len(keys) == 4


def test_generator(csv_connector: CsvConnector):
    for df in csv_connector.iter(chunksize=1):
        assert len(df) == 1
        assert df.columns.tolist() == ["index", "a", "b", "c"]

    for df in csv_connector.iter(offset=1, chunksize=1):
        assert len(df) == 1
        assert df.columns.tolist() == ["index", "a", "b", "c"]

    assert len(list(csv_connector.iter(chunksize=1))) == 2
    assert len(list(csv_connector.iter(offset=1, chunksize=1))) == 1


if __name__ == "__main__":
    pytest.main(["-vv", "-s", __file__])



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any

import pandas as pd

from sdgx.data_models.inspectors.base import RelationshipInspector
from sdgx.data_models.inspectors.extension import hookimpl
from sdgx.data_models.relationship import Relationship

if TYPE_CHECKING:
    from sdgx.data_models.metadata import Metadata


class SubsetRelationshipInspector(RelationshipInspector):
    """
    Inspecting relationships by comparing two columns is subset or not. So it needs to inspect all data for prev
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.maybe_related_columns: dict[str, dict[str, pd.Series]] = {}

    def _is_related(self, p: pd.Series, c: pd.Series) -> bool:
        """
        If child is subset of parent, assume related
        """

        return c.isin(p).all()

    def _build_relationship(self) -> list[Relationship]:
        r = []
        for parent, p_m_related in self.maybe_related_columns.items():
            for child, c_m_related in self.maybe_related_columns.items():
                if parent == child:
                    continue
                related_pairs = []
                for p_col, p_df in p_m_related.items():
                    for c_col, c_df in c_m_related.items():
                        if self._is_related(p_df, c_df):
                            related_pairs.append((p_col, c_col) if p_col != c_col else p_col)
                if related_pairs:
                    r.append(Relationship.build(parent, child, related_pairs))
        return r

    def fit(
        self,
        raw_data: pd.DataFrame,
        name: str | None = None,
        metadata: "Metadata" | None = None,
        *args,
        **kwargs,
    ):
        columns = set(n for n in chain(metadata.id_columns, metadata.primary_keys))
        for c in columns:
            cur_map = self.maybe_related_columns.setdefault(name, dict())
            cur_map[c] = pd.concat(
                (cur_map.get(c, pd.Series()), raw_data[[c]].squeeze()),
                ignore_index=True,
            )

    def inspect(self, *args, **kwargs) -> dict[str, Any]:
        """Inspect raw data and generate metadata."""
        return {"relationships": self._build_relationship()}


@hookimpl
def register(manager):
    manager.register("SubsetRelationshipInspector", SubsetRelationshipInspector)
